# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from storm.databases.sqlite import SQLite
from storm.exceptions import ConnectionBlockedError
from storm.store import Store, block_access
from storm.tests.helper import TestHelper, MakePath
from storm.uri import URI


class BlockAccessTest(TestHelper):
    """Tests for L{block_access}."""

    helpers = [MakePath]

    def setUp(self):
        super().setUp()
        database = SQLite(URI("sqlite:"))
        self.store = Store(database)

    def test_block_access(self):
        """
        The L{block_access} context manager blocks access to a L{Store}.  A
        L{ConnectionBlockedError} exception is raised if an attempt to access
        the underlying database is made while a store is blocked.
        """
        with block_access(self.store):
            self.assertRaises(ConnectionBlockedError, self.store.execute,
                              "SELECT 1")
        result = self.store.execute("SELECT 1")
        self.assertEqual([(1,)], list(result))

    def test_block_access_with_multiple_stores(self):
        """
        If multiple L{Store}s are passed to L{block_access} they will all be
        blocked until the managed context is left.
        """
        database = SQLite(URI("sqlite:%s" % self.make_path()))
        store = Store(database)
        with block_access(self.store, store):
            self.assertRaises(ConnectionBlockedError, self.store.execute,
                              "SELECT 1")
            self.assertRaises(ConnectionBlockedError, store.execute,
                              "SELECT 1")



#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import sys
import types
import gc

from storm.exceptions import ClosedError, DatabaseError, DisconnectionError
from storm.variables import Variable
import storm.database
from storm.database import *
from storm.tracer import install_tracer, remove_all_tracers, DebugTracer
from storm.uri import URI
from storm.expr import *
from storm.tests.helper import TestHelper
from storm.tests.mocker import ARGS


marker = object()


class RawConnection:

    closed = False

    def __init__(self, executed):
        self.executed = executed

    def cursor(self):
        return RawCursor(executed=self.executed)

    def commit(self):
        self.executed.append("COMMIT")

    def rollback(self):
        self.executed.append("ROLLBACK")

    def close(self):
        self.executed.append("CCLOSE")


class RawCursor:

    def __init__(self, arraysize=1, executed=None):
        self.arraysize = arraysize
        if executed is None:
            self.executed = []
        else:
            self.executed = executed

        self._fetchone_data = [("fetchone%d" % i,) for i in range(3)]
        self._fetchall_data = [("fetchall%d" % i,) for i in range(2)]
        self._fetchmany_data = [("fetchmany%d" % i,) for i in range(5)]

    def close(self):
        self.executed.append("RCLOSE")

    def execute(self, statement, params=marker):
        self.executed.append((statement, params))

    def fetchone(self):
        if self._fetchone_data:
            return self._fetchone_data.pop(0)
        return None

    def fetchall(self):
        result = self._fetchall_data
        self._fetchall_data = []
        return result

    def fetchmany(self):
        result = self._fetchmany_data[:self.arraysize]
        del self._fetchmany_data[:self.arraysize]
        return result


class FakeConnection:

    def __init__(self):
        self._database = Database()

    def _check_disconnect(self, _function, *args, **kwargs):
        return _function(*args, **kwargs)


class FakeTracer:

    def __init__(self, stream=None):
        self.seen = []

    def connection_raw_execute(self, connection, raw_cursor,
                               statement, params):
        self.seen.append(("EXECUTE", connection, type(raw_cursor),
                          statement, params))

    def connection_raw_execute_success(self, connection, raw_cursor,
                                       statement, params):
        self.seen.append(("SUCCESS", connection, type(raw_cursor),
                          statement, params))

    def connection_raw_execute_error(self, connection, raw_cursor,
                                     statement, params, error):
        self.seen.append(("ERROR", connection, type(raw_cursor),
                          statement, params, error))

    def connection_commit(self, connection, xid=None):
        self.seen.append(("COMMIT", connection, xid))

    def connection_rollback(self, connection, xid=None):
        self.seen.append(("ROLLBACK", connection, xid))


class DatabaseTest(TestHelper):

    def setUp(self):
        TestHelper.setUp(self)
        self.database = Database()

    def test_connect(self):
        self.assertRaises(NotImplementedError, self.database.connect)


class ConnectionTest(TestHelper):

    def setUp(self):
        TestHelper.setUp(self)
        self.executed = []
        self.database = Database()
        self.raw_connection = RawConnection(self.executed)
        self.database.raw_connect = lambda: self.raw_connection
        self.connection = Connection(self.database)

    def tearDown(self):
        TestHelper.tearDown(self)
        remove_all_tracers()

    def test_execute(self):
        result = self.connection.execute("something")
        self.assertTrue(isinstance(result, Result))
        self.assertEqual(self.executed, [("something", marker)])

    def test_execute_params(self):
        result = self.connection.execute("something", (1,2,3))
        self.assertTrue(isinstance(result, Result))
        self.assertEqual(self.executed, [("something", (1,2,3))])

    def test_execute_noresult(self):
        result = self.connection.execute("something", noresult=True)
        self.assertEqual(result, None)
        self.assertEqual(self.executed, [("something", marker), "RCLOSE"])

    def test_execute_convert_param_style(self):
        class MyConnection(Connection):
            param_mark = "%s"
        connection = MyConnection(self.database)
        result = connection.execute("'?' ? '?' ? '?'")
        self.assertEqual(self.executed, [("'?' %s '?' %s '?'", marker)])

        # TODO: Unsupported for now.
        #result = connection.execute("$$?$$ ? $asd$'?$asd$ ? '?'")
        #self.assertEqual(self.executed,
        #                 [("'?' %s '?' %s '?'", marker),
        #                  ("$$?$$ %s $asd'?$asd$ %s '?'", marker)])

    def test_execute_select(self):
        select = Select([SQLToken("column1"), SQLToken("column2")],
                        tables=[SQLToken("table1"), SQLToken("table2")])
        result = self.connection.execute(select)
        self.assertTrue(isinstance(result, Result))
        self.assertEqual(self.executed,
                         [("SELECT column1, column2 FROM table1, table2",
                           marker)])

    def test_execute_select_and_params(self):
        select = Select(["column1", "column2"], tables=["table1", "table2"])
        self.assertRaises(ValueError, self.connection.execute,
                          select, ("something",))

    def test_execute_closed(self):
        self.connection.close()
        self.assertRaises(ClosedError, self.connection.execute, "SELECT 1")

    def test_raw_execute_tracing(self):
        self.assertMethodsMatch(FakeTracer, DebugTracer)
        tracer = FakeTracer()
        install_tracer(tracer)
        self.connection.execute("something")
        self.assertEqual(tracer.seen, [("EXECUTE", self.connection, RawCursor,
                                        "something", ()),
                                       ("SUCCESS", self.connection, RawCursor,
                                        "something", ())])

        del tracer.seen[:]
        self.connection.execute("something", (1, 2))
        self.assertEqual(tracer.seen, [("EXECUTE", self.connection, RawCursor,
                                        "something", (1, 2)),
                                       ("SUCCESS", self.connection, RawCursor,
                                        "something", (1, 2))])

    def test_raw_execute_error_tracing(self):
        cursor_mock = self.mocker.patch(RawCursor)
        cursor_mock.execute(ARGS)
        error = ZeroDivisionError()
        self.mocker.throw(error)
        self.mocker.replay()

        self.assertMethodsMatch(FakeTracer, DebugTracer)
        tracer = FakeTracer()
        install_tracer(tracer)
        self.assertRaises(ZeroDivisionError,
                          self.connection.execute, "something")
        self.assertEqual(tracer.seen, [("EXECUTE", self.connection, RawCursor,
                                        "something", ()),
                                       ("ERROR", self.connection, RawCursor,
                                        "something", (), error)])

    def test_raw_execute_setup_error_tracing(self):
        """
        When an exception is raised in the connection_raw_execute hook of a
        tracer, the connection_raw_execute_error hook is called.
        """
        cursor_mock = self.mocker.patch(FakeTracer)
        cursor_mock.connection_raw_execute(ARGS)
        error = ZeroDivisionError()
        self.mocker.throw(error)
        self.mocker.replay()

        tracer = FakeTracer()
        install_tracer(tracer)
        self.assertRaises(ZeroDivisionError,
                          self.connection.execute, "something")
        self.assertEqual(tracer.seen, [("ERROR", self.connection, RawCursor,
                                        "something", (), error)])

    def test_tracing_check_disconnect(self):
        tracer = FakeTracer()
        tracer_mock = self.mocker.patch(tracer)
        tracer_mock.connection_raw_execute(ARGS)
        self.mocker.throw(DatabaseError('connection closed'))
        self.mocker.replay()

        install_tracer(tracer_mock)
        self.connection.is_disconnection_error = (
            lambda exc, extra_disconnection_errors=():
                'connection closed' in str(exc))

        self.assertRaises(DisconnectionError,
                          self.connection.execute, "something")

    def test_tracing_success_check_disconnect(self):
        tracer = FakeTracer()
        tracer_mock = self.mocker.patch(tracer)
        tracer_mock.connection_raw_execute(ARGS)
        tracer_mock.connection_raw_execute_success(ARGS)
        self.mocker.throw(DatabaseError('connection closed'))
        self.mocker.replay()

        install_tracer(tracer_mock)
        self.connection.is_disconnection_error = (
            lambda exc, extra_disconnection_errors=():
                'connection closed' in str(exc))

        self.assertRaises(DisconnectionError,
                          self.connection.execute, "something")

    def test_tracing_error_check_disconnect(self):
        cursor_mock = self.mocker.patch(RawCursor)
        cursor_mock.execute(ARGS)
        error = ZeroDivisionError()
        self.mocker.throw(error)
        tracer = FakeTracer()
        tracer_mock = self.mocker.patch(tracer)
        tracer_mock.connection_raw_execute(ARGS)
        tracer_mock.connection_raw_execute_error(ARGS)
        self.mocker.throw(DatabaseError('connection closed'))
        self.mocker.replay()

        install_tracer(tracer_mock)
        self.connection.is_disconnection_error = (
            lambda exc, extra_disconnection_errors=():
                'connection closed' in str(exc))

        self.assertRaises(DisconnectionError,
                          self.connection.execute, "something")

    def test_commit(self):
        self.connection.commit()
        self.assertEqual(self.executed, ["COMMIT"])

    def test_commit_tracing(self):
        self.assertMethodsMatch(FakeTracer, DebugTracer)
        tracer = FakeTracer()
        install_tracer(tracer)
        self.connection.commit()
        self.assertEqual(tracer.seen, [("COMMIT", self.connection, None)])

    def test_rollback(self):
        self.connection.rollback()
        self.assertEqual(self.executed, ["ROLLBACK"])

    def test_rollback_tracing(self):
        self.assertMethodsMatch(FakeTracer, DebugTracer)
        tracer = FakeTracer()
        install_tracer(tracer)
        self.connection.rollback()
        self.assertEqual(tracer.seen, [("ROLLBACK", self.connection, None)])

    def test_close(self):
        self.connection.close()
        self.assertEqual(self.executed, ["CCLOSE"])

    def test_close_twice(self):
        self.connection.close()
        self.connection.close()
        self.assertEqual(self.executed, ["CCLOSE"])

    def test_close_deallocates_raw_connection(self):
        refs_before = len(gc.get_referrers(self.raw_connection))
        self.connection.close()
        refs_after = len(gc.get_referrers(self.raw_connection))
        self.assertEqual(refs_after, refs_before-1)

    def test_del_deallocates_raw_connection(self):
        refs_before = len(gc.get_referrers(self.raw_connection))
        self.connection.__del__()
        refs_after = len(gc.get_referrers(self.raw_connection))
        self.assertEqual(refs_after, refs_before-1)

    def test_wb_del_with_previously_deallocated_connection(self):
        self.connection._raw_connection = None
        self.connection.__del__()

    def test_get_insert_identity(self):
        result = self.connection.execute("INSERT")
        self.assertRaises(NotImplementedError,
                          result.get_insert_identity, None, None)

    def test_wb_ensure_connected_noop(self):
        """Check that _ensure_connected() is a no-op for STATE_CONNECTED."""
        self.assertEqual(self.connection._state, storm.database.STATE_CONNECTED)
        def connect():
            raise DatabaseError("_ensure_connected() tried to connect")
        self.database.raw_connect = connect
        self.connection._ensure_connected()

    def test_wb_ensure_connected_dead_connection(self):
        """Check that DisconnectionError is raised for STATE_DISCONNECTED."""
        self.connection._state = storm.database.STATE_DISCONNECTED
        self.assertRaises(DisconnectionError,
                          self.connection._ensure_connected)

    def test_wb_ensure_connected_reconnects(self):
        """Check that _ensure_connected() reconnects for STATE_RECONNECT."""
        self.connection._state = storm.database.STATE_RECONNECT
        self.connection._raw_connection = None

        self.connection._ensure_connected()
        self.assertNotEqual(self.connection._raw_connection, None)
        self.assertEqual(self.connection._state,
                         storm.database.STATE_CONNECTED)

    def test_wb_ensure_connected_connect_failure(self):
        """Check that the connection is flagged on reconnect failures."""
        self.connection._state = storm.database.STATE_RECONNECT
        self.connection._raw_connection = None
        def _fail_to_connect():
            raise DatabaseError("could not connect")
        self.database.raw_connect = _fail_to_connect

        self.assertRaises(DisconnectionError,
                          self.connection._ensure_connected)
        self.assertEqual(self.connection._state,
                         storm.database.STATE_DISCONNECTED)
        self.assertEqual(self.connection._raw_connection, None)

    def test_wb_check_disconnection(self):
        """Ensure that _check_disconnect() detects disconnections."""
        class FakeException(DatabaseError):
            """A fake database exception that indicates a disconnection."""
        self.connection.is_disconnection_error = (
            lambda exc, extra_disconnection_errors=():
                isinstance(exc, FakeException))

        self.assertEqual(self.connection._state,
                         storm.database.STATE_CONNECTED)
        # Error is converted to DisconnectionError:
        def raise_exception():
            raise FakeException
        self.assertRaises(DisconnectionError,
                          self.connection._check_disconnect, raise_exception)
        self.assertEqual(self.connection._state,
                         storm.database.STATE_DISCONNECTED)
        self.assertEqual(self.connection._raw_connection, None)

    def test_wb_check_disconnection_extra_errors(self):
        """Ensure that _check_disconnect() can check for additional
        exceptions."""
        class FakeException(DatabaseError):
            """A fake database exception that indicates a disconnection."""
        self.connection.is_disconnection_error = (
            lambda exc, extra_disconnection_errors=():
                isinstance(exc, extra_disconnection_errors))

        self.assertEqual(self.connection._state,
                         storm.database.STATE_CONNECTED)
        # Error is converted to DisconnectionError:
        def raise_exception():
            raise FakeException
        # Exception passes through as normal.
        self.assertRaises(FakeException,
                          self.connection._check_disconnect, raise_exception)
        self.assertEqual(self.connection._state,
                         storm.database.STATE_CONNECTED)
        # Exception treated as a disconnection when keyword argument passed.
        self.assertRaises(DisconnectionError,
                          self.connection._check_disconnect, raise_exception,
                          extra_disconnection_errors=FakeException)
        self.assertEqual(self.connection._state,
                         storm.database.STATE_DISCONNECTED)

    def test_wb_rollback_clears_disconnected_connection(self):
        """Check that rollback clears the DISCONNECTED state."""
        self.connection._state = storm.database.STATE_DISCONNECTED
        self.connection._raw_connection = None

        self.connection.rollback()
        self.assertEqual(self.executed, [])
        self.assertEqual(self.connection._state,
                         storm.database.STATE_RECONNECT)


class ResultTest(TestHelper):

    def setUp(self):
        TestHelper.setUp(self)
        self.executed = []
        self.raw_cursor = RawCursor(executed=self.executed)
        self.result = Result(FakeConnection(), self.raw_cursor)

    def test_get_one(self):
        self.assertEqual(self.result.get_one(), ("fetchone0",))
        self.assertEqual(self.result.get_one(), ("fetchone1",))
        self.assertEqual(self.result.get_one(), ("fetchone2",))
        self.assertEqual(self.result.get_one(), None)

    def test_get_all(self):
        self.assertEqual(self.result.get_all(),
                         [("fetchall0",), ("fetchall1",)])
        self.assertEqual(self.result.get_all(), [])

    def test_iter(self):
        result = Result(FakeConnection(), RawCursor(2))
        self.assertEqual([item for item in result],
                         [("fetchmany0",), ("fetchmany1",), ("fetchmany2",),
                          ("fetchmany3",), ("fetchmany4",)])

    def test_set_variable(self):
        variable = Variable()
        self.result.set_variable(variable, marker)
        self.assertEqual(variable.get(), marker)

    def test_close(self):
        self.result.close()
        self.assertEqual(self.executed, ["RCLOSE"])

    def test_close_twice(self):
        self.result.close()
        self.result.close()
        self.assertEqual(self.executed, ["RCLOSE"])

    def test_close_deallocates_raw_cursor(self):
        refs_before = len(gc.get_referrers(self.raw_cursor))
        self.result.close()
        refs_after = len(gc.get_referrers(self.raw_cursor))
        self.assertEqual(refs_after, refs_before-1)

    def test_del_deallocates_raw_cursor(self):
        refs_before = len(gc.get_referrers(self.raw_cursor))
        self.result.__del__()
        refs_after = len(gc.get_referrers(self.raw_cursor))
        self.assertEqual(refs_after, refs_before-1)

    def test_wb_del_with_previously_deallocated_cursor(self):
        self.result._raw_cursor = None
        self.result.__del__()

    def test_set_arraysize(self):
        """When the arraysize is 1, change it to a better value."""
        raw_cursor = RawCursor()
        self.assertEqual(raw_cursor.arraysize, 1)
        result = Result(FakeConnection(), raw_cursor)
        self.assertEqual(raw_cursor.arraysize, 10)

    def test_preserve_arraysize(self):
        """When the arraysize is not 1, preserve it."""
        raw_cursor = RawCursor(arraysize=123)
        result = Result(FakeConnection(), raw_cursor)
        self.assertEqual(raw_cursor.arraysize, 123)


class CreateDatabaseTest(TestHelper):

    def setUp(self):
        TestHelper.setUp(self)
        self.db_module = types.ModuleType("db_module")
        self.uri = None
        def create_from_uri(uri):
            self.uri = uri
            return "RESULT"
        self.db_module.create_from_uri = create_from_uri
        sys.modules["storm.databases.db_module"] = self.db_module

    def tearDown(self):
        del sys.modules["storm.databases.db_module"]
        TestHelper.tearDown(self)

    def test_create_database_with_str(self):
        create_database("db_module:db")
        self.assertTrue(self.uri)
        self.assertEqual(self.uri.scheme, "db_module")
        self.assertEqual(self.uri.database, "db")

    def test_create_database_with_unicode(self):
        create_database("db_module:db")
        self.assertTrue(self.uri)
        self.assertEqual(self.uri.scheme, "db_module")
        self.assertEqual(self.uri.database, "db")

    def test_create_database_with_uri(self):
        uri = URI("db_module:db")
        create_database(uri)
        self.assertTrue(self.uri is uri)


class RegisterSchemeTest(TestHelper):

    uri = None

    def tearDown(self):
        if 'factory' in storm.database._database_schemes:
            del storm.database._database_schemes['factory']
        TestHelper.tearDown(self)

    def test_register_scheme(self):
        def factory(uri):
            self.uri = uri
            return "FACTORY RESULT"
        register_scheme('factory', factory)
        self.assertEqual(storm.database._database_schemes['factory'], factory)
        # Check that we can create databases that use this scheme ...
        result = create_database('factory:foobar')
        self.assertEqual(result, "FACTORY RESULT")
        self.assertTrue(self.uri)
        self.assertEqual(self.uri.scheme, 'factory')
        self.assertEqual(self.uri.database, 'foobar')



#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from bisect import insort_left, bisect_left
import weakref
import sys

from storm.exceptions import PropertyPathError
from storm.info import get_obj_info, get_cls_info
from storm.expr import Column, Undef
from storm.variables import (
    Variable, VariableFactory, BoolVariable, IntVariable, FloatVariable,
    DecimalVariable, BytesVariable, UnicodeVariable, DateTimeVariable,
    DateVariable, TimeVariable, TimeDeltaVariable, UUIDVariable,
    PickleVariable, JSONVariable, ListVariable, EnumVariable)



__all__ = ["Property", "SimpleProperty",
           "Bool", "Int", "Float", "Decimal", "Bytes", "RawStr", "Unicode",
           "DateTime", "Date", "Time", "TimeDelta", "UUID", "Enum",
           "Pickle", "JSON", "List", "PropertyRegistry"]


class Property:
    """A property representing a database column.

    Properties can be set as attributes of classes that have a
    C{__storm_table__}, and can then be used like ordinary Python properties
    on instances of the class, corresponding to database columns.
    """

    def __init__(self, name=None, primary=False,
                 variable_class=Variable, variable_kwargs={}):
        """
        @param name: The name of this property.
        @param primary: A boolean indicating whether this property is a
            primary key.
        @param variable_class: The type of L{storm.variables.Variable}
            corresponding to this property.
        @param variable_kwargs: Dictionary of keyword arguments to be passed
            when constructing the underlying variable.
        """
        self._name = name
        self._primary = primary
        self._variable_class = variable_class
        self._variable_kwargs = variable_kwargs

    def __get__(self, obj, cls=None):
        if obj is None:
            return self._get_column(cls)
        obj_info = get_obj_info(obj)
        if cls is None:
            # Don't get obj.__class__ because we don't trust it
            # (might be proxied or whatever).
            cls = obj_info.cls_info.cls
        column = self._get_column(cls)
        return obj_info.variables[column].get()

    def __set__(self, obj, value):
        obj_info = get_obj_info(obj)
        # Don't get obj.__class__ because we don't trust it
        # (might be proxied or whatever).
        column = self._get_column(obj_info.cls_info.cls)
        obj_info.variables[column].set(value)

    def __delete__(self, obj):
        obj_info = get_obj_info(obj)
        # Don't get obj.__class__ because we don't trust it
        # (might be proxied or whatever).
        column = self._get_column(obj_info.cls_info.cls)
        obj_info.variables[column].delete()

    def _detect_attr_name(self, used_cls):
        self_id = id(self)
        for cls in used_cls.__mro__:
            for attr, prop in cls.__dict__.items():
                if id(prop) == self_id:
                    return attr
        raise RuntimeError("Property used in an unknown class")

    def _get_column(self, cls):
        # Cache per-class column values in the class itself, to avoid
        # holding a strong reference to it here, and thus rendering
        # classes uncollectable in certain situations (e.g. subclasses
        # where the property is stored in the base).
        try:
            # Use class dictionary explicitly to get sensible
            # results on subclasses.
            column = cls.__dict__["_storm_columns"].get(self)
        except KeyError:
            cls._storm_columns = {}
            column = None
        if column is None:
            attr = self._detect_attr_name(cls)
            if self._name is None:
                name = attr
            else:
                name = self._name
            column = PropertyColumn(self, cls, attr, name, self._primary,
                                    self._variable_class,
                                    self._variable_kwargs)
            cls._storm_columns[self] = column
        return column


class PropertyColumn(Column):

    def __init__(self, prop, cls, attr, name, primary,
                 variable_class, variable_kwargs):
        Column.__init__(self, name, cls, primary,
                        VariableFactory(variable_class, column=self,
                                        validator_attribute=attr,
                                        **variable_kwargs))

        self.cls = cls # Used by references

        # Copy attributes from the property to avoid one additional
        # function call on each access.
        for attr in ["__get__", "__set__", "__delete__"]:
            setattr(self, attr, getattr(prop, attr))


class SimpleProperty(Property):

    variable_class = None

    def __init__(self, name=None, primary=False, **kwargs):
        """
        @param name: The name of this property.
        @param primary: A boolean indicating whether this property is a
            primary key.
        @param default: The initial value of this variable.  The default
            behavior is for the value to stay undefined until it is
            set with L{set}.
        @param default_factory: If specified, this will immediately be
            called to get the initial value.
        @param allow_none: A boolean indicating whether None should be
            allowed to be set as the value of this variable.
        @param validator: Validation function called whenever trying to
            set the variable to a non-db value.  The function should
            look like validator(object, attr, value), where the first and
            second arguments are the result of validator_object_factory()
            (or None, if this parameter isn't provided) and the value of
            validator_attribute, respectively.  When called, the function
            should raise an error if the value is unacceptable, or return
            the value to be used in place of the original value otherwise.
        @param kwargs: Other keyword arguments passed through when
            constructing the underlying variable.
        """
        kwargs["value"] = kwargs.pop("default", Undef)
        kwargs["value_factory"] = kwargs.pop("default_factory", Undef)
        Property.__init__(self, name, primary, self.variable_class, kwargs)


class Bool(SimpleProperty):
    """Boolean property.

    This accepts integer, L{float}, or L{decimal.Decimal} values, and stores
    them as booleans.
    """
    variable_class = BoolVariable


class Int(SimpleProperty):
    """Integer property.

    This accepts integer, L{float}, or L{decimal.Decimal} values, and stores
    them as integers.
    """
    variable_class = IntVariable


class Float(SimpleProperty):
    """Float property.

    This accepts integer, L{float}, or L{decimal.Decimal} values, and stores
    them as floating-point values.
    """
    variable_class = FloatVariable


class Decimal(SimpleProperty):
    """Decimal property.

    This accepts integer or L{decimal.Decimal} values, and stores them as
    text strings containing their decimal representation.
    """
    variable_class = DecimalVariable


class Bytes(SimpleProperty):
    """Bytes property.

    This accepts L{bytes} or L{memoryview} objects, and stores them as byte
    strings.

    Deprecated aliases: L{Chars}, L{RawStr}.
    """
    variable_class = BytesVariable


# OBSOLETE: Bytes was Chars in 0.9. This will die soon.
Chars = Bytes
# DEPRECATED: Bytes was RawStr until 0.22.
RawStr = Bytes


class Unicode(SimpleProperty):
    """Unicode property.

    This accepts L{str} objects, and stores them as text strings.
    """
    variable_class = UnicodeVariable


class DateTime(SimpleProperty):
    """Date and time property.

    This accepts aware L{datetime.datetime} objects and stores them as
    timestamps; it also accepts integer or L{float} objects, converting them
    using L{datetime.utcfromtimestamp}.  Note that it does not accept naive
    L{datetime.datetime} objects (those that do not have timezone
    information).
    """
    variable_class = DateTimeVariable


class Date(SimpleProperty):
    """Date property.

    This accepts L{datetime.date} objects and stores them as datestamps; it
    also accepts L{datetime.datetime} objects, converting them using
    L{datetime.datetime.date}.
    """
    variable_class = DateVariable


class Time(SimpleProperty):
    """Time property.

    This accepts L{datetime.time} objects and stores them as datestamps; it
    also accepts L{datetime.datetime} objects, converting them using
    L{datetime.datetime.time}.
    """
    variable_class = TimeVariable


class TimeDelta(SimpleProperty):
    """Time delta property.

    This accepts L{datetime.timedelta} objects and stores them as time
    intervals.
    """
    variable_class = TimeDeltaVariable


class UUID(SimpleProperty):
    """UUID property.

    This accepts L{uuid.UUID} objects and stores them as their text
    representation.
    """
    variable_class = UUIDVariable


class Pickle(SimpleProperty):
    """Pickle property.

    This accepts any object that can be serialized using L{pickle}, and
    stores it as a byte string containing its pickled representation.
    """
    variable_class = PickleVariable


class JSON(SimpleProperty):
    """JSON property.

    This accepts any object that can be serialized using L{json}, and stores
    it as a text string containing its JSON representation.
    """
    variable_class = JSONVariable


class List(SimpleProperty):
    """List property.

    This accepts iterable objects and stores them as a list where each
    element is an object of the given value type.
    """
    variable_class = ListVariable

    def __init__(self, name=None, **kwargs):
        """
        @param name: The name of this property.
        @param type: An instance of L{Property} defining the type of each
            element of this list.
        @param default_factory: If specified, this will immediately be
            called to get the initial value.
        @param validator: Validation function called whenever trying to
            set the variable to a non-db value.  The function should
            look like validator(object, attr, value), where the first and
            second arguments are the result of validator_object_factory()
            (or None, if this parameter isn't provided) and the value of
            validator_attribute, respectively.  When called, the function
            should raise an error if the value is unacceptable, or return
            the value to be used in place of the original value otherwise.
        @param kwargs: Other keyword arguments passed through when
            constructing the underlying variable.
        """
        if "default" in kwargs:
            raise ValueError("'default' not allowed for List. "
                             "Use 'default_factory' instead.")
        type = kwargs.pop("type", None)
        if type is None:
            type = Property()
        kwargs["item_factory"] = VariableFactory(type._variable_class,
                                                 **type._variable_kwargs)
        SimpleProperty.__init__(self, name, **kwargs)


class Enum(SimpleProperty):
    """Enumeration property, allowing used values to differ from stored ones.

    For instance::

        class Class(Storm):
            prop = Enum(map={"one": 1, "two": 2})

        obj.prop = "one"
        assert obj.prop == "one"

        obj.prop = 1 # Raises error.

    Another example::

        class Class(Storm):
            prop = Enum(map={"one": 1, "two": 2}, set_map={"um": 1})

        obj.prop = "um"
        assert obj.prop is "one"

        obj.prop = "one" # Raises error.
    """
    variable_class = EnumVariable

    def __init__(self, name=None, primary=False, **kwargs):
        set_map = dict(kwargs.pop("map"))
        get_map = {value: key for key, value in set_map.items()}
        if "set_map" in kwargs:
            set_map = dict(kwargs.pop("set_map"))

        kwargs["get_map"] = get_map
        kwargs["set_map"] = set_map
        SimpleProperty.__init__(self, name, primary, **kwargs)


class PropertyRegistry:
    """
    An object which remembers the Storm properties specified on
    classes, and is able to translate names to these properties.
    """
    def __init__(self):
        self._properties = []

    def get(self, name, namespace=None):
        """Translate a property name path to the actual property.

        This method accepts a property name like C{"id"} or C{"Class.id"}
        or C{"module.path.Class.id"}, and tries to find a unique
        class/property with the given name.

        When the C{namespace} argument is given, the registry will be
        able to disambiguate names by choosing the one that is closer
        to the given namespace.  For instance C{get("Class.id", "a.b.c")}
        will choose C{a.Class.id} rather than C{d.Class.id}.
        """
        key = ".".join(reversed(name.split(".")))+"."
        i = bisect_left(self._properties, (key,))
        l = len(self._properties)
        best_props = []
        if namespace is None:
            while i < l and self._properties[i][0].startswith(key):
                path, prop_ref = self._properties[i]
                prop = prop_ref()
                if prop is not None:
                    best_props.append((path, prop))
                i += 1
        else:
            namespace_parts = ("." + namespace).split(".")
            best_path_info = (0, sys.maxsize)
            while i < l and self._properties[i][0].startswith(key):
                path, prop_ref = self._properties[i]
                prop = prop_ref()
                if prop is None:
                    i += 1
                    continue
                path_parts = path.split(".")
                path_parts.reverse()
                common_prefix = 0
                for part, ns_part in zip(path_parts, namespace_parts):
                    if part == ns_part:
                        common_prefix += 1
                    else:
                        break
                path_info = (-common_prefix, len(path_parts)-common_prefix)
                if path_info < best_path_info:
                    best_path_info = path_info
                    best_props = [(path, prop)]
                elif path_info == best_path_info:
                    best_props.append((path, prop))
                i += 1
        if not best_props:
            raise PropertyPathError("Path '%s' matches no known property."
                                    % name)
        elif len(best_props) > 1:
            paths = [".".join(reversed(path.split(".")[:-1]))
                     for path, prop in best_props]
            raise PropertyPathError("Path '%s' matches multiple "
                                    "properties: %s" %
                                    (name, ", ".join(paths)))
        return best_props[0][1]

    def add_class(self, cls):
        """Register properties of C{cls} so that they may be found by C{get()}.
        """
        suffix = cls.__module__.split(".")
        suffix.append(cls.__name__)
        suffix.reverse()
        suffix = ".%s." % ".".join(suffix)
        cls_info = get_cls_info(cls)
        for attr in cls_info.attributes:
            prop = cls_info.attributes[attr]
            prop_ref = weakref.KeyedRef(prop, self._remove, None)
            pair = (attr+suffix, prop_ref)
            prop_ref.key = pair
            insort_left(self._properties, pair)

    def add_property(self, cls, prop, attr_name):
        """Register property of C{cls} so that it may be found by C{get()}.
        """
        suffix = cls.__module__.split(".")
        suffix.append(cls.__name__)
        suffix.reverse()
        suffix = ".%s." % ".".join(suffix)
        prop_ref = weakref.KeyedRef(prop, self._remove, None)
        pair = (attr_name+suffix, prop_ref)
        prop_ref.key = pair
        insort_left(self._properties, pair)

    def clear(self):
        """Clean up all properties in the registry.

        Used by tests.
        """
        del self._properties[:]

    def _remove(self, ref):
        self._properties.remove(ref.key)


class PropertyPublisherMeta(type):
    """A metaclass that associates subclasses with Storm L{PropertyRegistry}s.
    """

    def __init__(self, name, bases, dict):
        if not hasattr(self, "_storm_property_registry"):
            self._storm_property_registry = PropertyRegistry()
        elif hasattr(self, "__storm_table__"):
            self._storm_property_registry.add_class(self)



#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Apply database patches.

The L{PatchApplier} class can be used to apply and keep track of a series
of database patches.

To create a patch series all is needed is to add Python files under a module
of choice, an name them as 'patch_N.py' where 'N' is the version of the patch
in the series. Each patch file must define an C{apply} callable taking a
L{Store} instance has its only argument. This function will be called when the
patch gets applied.

The L{PatchApplier} can be then used to apply to a L{Store} all the available
patches. After a patch has been applied, its version is recorded in a special
'patch' table in the given L{Store}, and it won't be applied again.
"""

import sys
import os
import re
import types

from storm.locals import StormError, Int


class UnknownPatchError(Exception):
    """
    Raised if a patch is found in the database that doesn't exist
    in the local patch directory.
    """

    def __init__(self, store, patches):
        self._store = store
        self._patches = patches

    def __str__(self):
        return "store has patches the code doesn't know about: %s" % (
            ", ".join([str(version) for version in self._patches]))


class BadPatchError(Exception):
    """Raised when a patch failing with a random exception is found."""


class Patch:
    """Database object representing an applied patch.

    @version: The version of the patch associated with this object.
    """

    __storm_table__ = "patch"

    version = Int(primary=True, allow_none=False)

    def __init__(self, version):
        self.version = version


class PatchApplier:
    """Apply to a L{Store} the database patches from a given Python package.

    @param store: The L{Store} to apply the patches to.
    @param patch_set: The L{PatchSet} containing the patches to apply.
    @param committer: Optionally an object implementing 'commit()' and
        'rollback()' methods, to be used to commit or rollback the changes
        after applying a patch. If C{None} is given, the C{store} itself is
        used.
    """

    def __init__(self, store, patch_set, committer=None):
        self._store = store
        if isinstance(patch_set, types.ModuleType):
            # Up to version 0.20.0 the second positional parameter used to
            # be a raw module containing the patches. We wrap it with PatchSet
            # for keeping backward-compatibility.
            patch_set = PatchSet(patch_set)
        self._patch_set = patch_set
        if committer is None:
            committer = store
        self._committer = committer

    def apply(self, version):
        """Execute the patch with the given version.

        This will call the 'apply' function defined in the patch file with
        the given version, passing it our L{Store}.

        @param version: The version of the patch to execute.
        """
        patch = Patch(version)
        self._store.add(patch)
        module = None
        try:
            module = self._patch_set.get_patch_module(version)
            module.apply(self._store)
        except StormError:
            self._committer.rollback()
            raise
        except:
            type, value, traceback = sys.exc_info()
            patch_repr = getattr(module, "__file__", version)
            raise BadPatchError(
                "Patch %s failed: %s: %s" %
                (patch_repr, type.__name__, str(value))
            ).with_traceback(traceback)
        self._committer.commit()

    def apply_all(self):
        """Execute all unapplied patches.

        @raises UnknownPatchError: If the patch table has versions for which
            no patch file actually exists.
        """
        self.check_unknown()
        for version in self.get_unapplied_versions():
            self.apply(version)

    def mark_applied(self, version):
        """Mark the patch with the given version as applied."""
        self._store.add(Patch(version))
        self._committer.commit()

    def mark_applied_all(self):
        """Mark all unapplied patches as applied."""
        for version in self.get_unapplied_versions():
            self.mark_applied(version)

    def has_pending_patches(self):
        """Return C{True} if there are unapplied patches, C{False} if not."""
        for version in self.get_unapplied_versions():
            return True
        return False

    def get_unknown_patch_versions(self):
        """
        Return the list of Patch versions that have been applied to the
        database, but don't appear in the schema's patches module.
        """
        applied = self._get_applied_patches()
        known_patches = self._patch_set.get_patch_versions()
        unknown_patches = set()

        for patch in applied:
            if not patch in known_patches:
                unknown_patches.add(patch)
        return unknown_patches

    def check_unknown(self):
        """Look for patches that we don't know about.

        @raises UnknownPatchError: If the store has applied patch versions
            this schema doesn't know about.
        """
        unknown_patches = self.get_unknown_patch_versions()
        if unknown_patches:
            raise UnknownPatchError(self._store, unknown_patches)

    def get_unapplied_versions(self):
        """Return the versions of all unapplied patches."""
        applied = self._get_applied_patches()
        for version in self._patch_set.get_patch_versions():
            if version not in applied:
                yield version

    def _get_applied_patches(self):
        """Return the versions of all applied patches."""
        applied = set()
        for patch in self._store.find(Patch):
            applied.add(patch.version)
        return applied


class PatchSet:
    """A collection of patch modules.

    Each patch module lives in a regular Python module file, contained in a
    sub-directory named against the patch version. For example, given
    a directory tree like:

      mypackage/
          __init__.py
          patch_1/
              __init__.py
              foo.py

    the following code will return a patch module object for foo.py:

      >>> import mypackage
      >>> patch_set = PackagePackage(mypackage, sub_level="foo")
      >>> patch_module = patch_set.get_patch_module(1)
      >>> print(patch_module.__name__)
      'mypackage.patch_1.foo'

    Different sub-levels can be used to apply different patches to different
    stores (see L{Sharding}).

    Alternatively if no sub-level is provided, the structure will be flat:

      mypackage/
          __init__.py
          patch_1.py

      >>> import mypackage
      >>> patch_set = PackagePackage(mypackage)
      >>> patch_module = patch_set.get_patch_module(1)
      >>> print(patch_module.__name__)
      'mypackage.patch_1'

    This simpler structure can be used if you have just one store to patch
    or you don't care to co-ordinate the patches across your stores.
    """

    def __init__(self, package, sub_level=None):
        self._package = package
        self._sub_level = sub_level

    def get_patch_versions(self):
        """Return the versions of all available patches."""
        pattern = r"^patch_(\d+)"
        if not self._sub_level:
            pattern += ".py"
        pattern += "$"
        format = re.compile(pattern)

        patch_directory = self._get_patch_directory()
        filenames = os.listdir(patch_directory)
        matches = [(format.match(fn), fn) for fn in filenames]
        matches = sorted(filter(lambda x: x[0], matches),
                         key=lambda x: int(x[0].group(1)))
        return [int(match.group(1)) for match, filename in matches]

    def get_patch_module(self, version):
        """Import the Python module of the patch file with the given version.

        @param: The version of the module patch to import.
        @return: The imported module.
        """
        name = "patch_%d" % version
        levels = [self._package.__name__, name]
        if self._sub_level:
            directory = self._get_patch_directory()
            path = os.path.join(directory, name, self._sub_level + ".py")
            if not os.path.exists(path):
                return _EmptyPatchModule()
            levels.append(self._sub_level)
        return __import__(".".join(levels), None, None, [''])

    def _get_patch_directory(self):
        """Get the path to the directory of the patch package."""
        return os.path.dirname(self._package.__file__)


class _EmptyPatchModule:
    """Fake module object with a no-op C{apply} function."""

    def apply(self, store):
        pass



#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import threading
import weakref
import gc

from storm.tests.helper import TestHelper
from storm.tests.zope import has_transaction, has_zope_component

if has_transaction:
    import transaction
    from transaction import ThreadTransactionManager
    from storm.zope.interfaces import IZStorm, ZStormError
    from storm.zope.zstorm import ZStorm, StoreDataManager

if has_zope_component:
    from zope.component import provideUtility, getUtility

from storm.exceptions import OperationalError
from storm.locals import Store


class ZStormTest(TestHelper):

    def is_supported(self):
        return has_transaction

    def setUp(self):
        self.zstorm = ZStorm()

    def tearDown(self):
        # Reset the utility to cleanup the StoreSynchronizer's from the
        # transaction.
        self.zstorm._reset()
        # Free the transaction to avoid having errors that cross
        # test cases.
        # XXX cjwatson 2019-05-29: transaction 2.4.0 changed
        # ThreadTransactionManager to wrap TransactionManager rather than
        # inheriting from it.  For now, cope with either.  Simplify this
        # once transaction 2.4.0 is old enough that we can reasonably just
        # test-depend on it.
        manager = transaction.manager
        if isinstance(manager, ThreadTransactionManager):
            try:
                manager.free
            except AttributeError:
                # transaction >= 2.4.0
                manager = manager.manager
        manager.free(transaction.get())

    def test_create(self):
        store = self.zstorm.create(None, "sqlite:")
        self.assertTrue(isinstance(store, Store))

    def test_create_twice_unnamed(self):
        store = self.zstorm.create(None, "sqlite:")
        store.execute("CREATE TABLE test (id INTEGER)")
        store.commit()

        store = self.zstorm.create(None, "sqlite:")
        self.assertRaises(OperationalError,
                          store.execute, "SELECT * FROM test")

    def test_create_twice_same_name(self):
        store = self.zstorm.create("name", "sqlite:")
        self.assertRaises(ZStormError, self.zstorm.create, "name", "sqlite:")

    def test_create_and_get_named(self):
        store = self.zstorm.create("name", "sqlite:")
        self.assertTrue(self.zstorm.get("name") is store)

    def test_create_and_get_named_another_thread(self):
        store = self.zstorm.create("name", "sqlite:")

        raised = []

        def f():
            try:
                self.zstorm.get("name")
            except ZStormError:
                raised.append(True)
        thread = threading.Thread(target=f)
        thread.start()
        thread.join()

        self.assertTrue(raised)

    def test_get_unexistent(self):
        self.assertRaises(ZStormError, self.zstorm.get, "name")

    def test_get_with_uri(self):
        store = self.zstorm.get("name", "sqlite:")
        self.assertTrue(isinstance(store, Store))
        self.assertTrue(self.zstorm.get("name") is store)
        self.assertTrue(self.zstorm.get("name", "sqlite:") is store)

    def test_set_default_uri(self):
        self.zstorm.set_default_uri("name", "sqlite:")
        store = self.zstorm.get("name")
        self.assertTrue(isinstance(store, Store))

    def test_create_default(self):
        self.zstorm.set_default_uri("name", "sqlite:")
        store = self.zstorm.create("name")
        self.assertTrue(isinstance(store, Store))

    def test_create_default_twice(self):
        self.zstorm.set_default_uri("name", "sqlite:")
        self.zstorm.create("name")
        self.assertRaises(ZStormError, self.zstorm.create, "name")

    def test_iterstores(self):
        store1 = self.zstorm.create(None, "sqlite:")
        store2 = self.zstorm.create(None, "sqlite:")
        store3 = self.zstorm.create("name", "sqlite:")
        stores = []
        for name, store in self.zstorm.iterstores():
            stores.append((name, store))
        self.assertEqual(len(stores), 3)
        self.assertEqual(set(stores),
                         {(None, store1), (None, store2), ("name", store3)})

    def test_get_name(self):
        store = self.zstorm.create("name", "sqlite:")
        self.assertEqual(self.zstorm.get_name(store), "name")

    def test_get_name_with_removed_store(self):
        store = self.zstorm.create("name", "sqlite:")
        self.assertEqual(self.zstorm.get_name(store), "name")
        self.zstorm.remove(store)
        self.assertEqual(self.zstorm.get_name(store), None)

    def test_default_databases(self):
        self.zstorm.set_default_uri("name1", "sqlite:1")
        self.zstorm.set_default_uri("name2", "sqlite:2")
        self.zstorm.set_default_uri("name3", "sqlite:3")
        default_uris = self.zstorm.get_default_uris()
        self.assertEqual(default_uris, {"name1": "sqlite:1",
                                        "name2": "sqlite:2",
                                        "name3": "sqlite:3"})

    def test_register_store_for_tpc_transaction(self):
        """
        Setting a store to use two-phase-commit mode, makes ZStorm
        call its begin() method when it joins the transaction.
        """
        self.zstorm.set_default_uri("name", "sqlite:")
        self.zstorm.set_default_tpc("name", True)
        store = self.zstorm.get("name")
        xids = []
        store.begin = lambda xid: xids.append(xid)
        store.execute("SELECT 1")
        [xid] = xids
        self.assertEqual(0, xid.format_id)
        self.assertEqual("_storm", xid.global_transaction_id[:6])
        self.assertEqual("name", xid.branch_qualifier)

    def test_register_store_for_tpc_transaction_uses_per_transaction_id(self):
        """
        Two stores in two-phase-commit mode joining the same transaction share
        the same global transaction ID.
        """
        self.zstorm.set_default_uri("name1", "sqlite:///%s" % self.makeFile())
        self.zstorm.set_default_uri("name2", "sqlite:///%s" % self.makeFile())
        self.zstorm.set_default_tpc("name1", True)
        self.zstorm.set_default_tpc("name2", True)
        store1 = self.zstorm.get("name1")
        store2 = self.zstorm.get("name2")
        xids = []
        store1.begin = lambda xid: xids.append(xid)
        store2.begin = lambda xid: xids.append(xid)
        store1.execute("SELECT 1")
        store2.execute("SELECT 1")
        [xid1, xid2] = xids
        self.assertEqual(xid1.global_transaction_id,
                         xid2.global_transaction_id)

    def test_register_store_for_tpc_transaction_uses_unique_global_ids(self):
        """
        Each global transaction gets assigned a unique ID.
        """
        self.zstorm.set_default_uri("name", "sqlite:")
        self.zstorm.set_default_tpc("name", True)
        store = self.zstorm.get("name")
        xids = []
        store.begin = lambda xid: xids.append(xid)
        store.execute("SELECT 1")
        transaction.abort()
        store.execute("SELECT 1")
        transaction.abort()
        [xid1, xid2] = xids
        self.assertNotEqual(xid1.global_transaction_id,
                            xid2.global_transaction_id)

    def test_transaction_with_two_phase_commit(self):
        """
        If a store is set to use TPC, than the associated data manager will
        call its prepare() and commit() methods when committing.
        """
        self.zstorm.set_default_uri("name", "sqlite:")
        self.zstorm.set_default_tpc("name", True)
        store = self.zstorm.get("name")
        calls = []
        store.begin = lambda xid: calls.append("begin")
        store.prepare = lambda: calls.append("prepare")
        store.commit = lambda: calls.append("commit")
        store.execute("SELECT 1")
        transaction.commit()
        self.assertEqual(["begin", "prepare", "commit"], calls)

    def test_transaction_with_single_and_two_phase_commit_stores(self):
        """
        When there are both stores in single-phase and two-phase mode, the
        ones in single-phase mode are committed first. This makes it possible
        to actually achieve two-phase commit behavior when only one store
        doesn't support TPC.
        """
        self.zstorm.set_default_uri("name1", "sqlite:///%s" % self.makeFile())
        self.zstorm.set_default_uri("name2", "sqlite:///%s" % self.makeFile())
        self.zstorm.set_default_tpc("name1", True)
        self.zstorm.set_default_tpc("name2", False)
        store1 = self.zstorm.get("name1")
        store2 = self.zstorm.get("name2")
        commits = []
        store1.begin = lambda xid: None
        store1.prepare = lambda: None
        store1.commit = lambda: commits.append("commit1")
        store2.commit = lambda: commits.append("commit2")
        store1.execute("SELECT 1")
        store2.execute("SELECT 1")
        transaction.commit()
        self.assertEqual(["commit2", "commit1"], commits)

    def _isInTransaction(self, store):
        """Check if a Store is part of the current transaction."""
        for dm in transaction.get()._resources:
            if isinstance(dm, StoreDataManager) and dm._store is store:
                return True
        return False

    def assertInTransaction(self, store):
        """Check that the given store is joined to the transaction."""
        self.assertTrue(self._isInTransaction(store),
                        "%r should be joined to the transaction" % store)

    def assertNotInTransaction(self, store):
        """Check that the given store is not joined to the transaction."""
        self.assertTrue(not self._isInTransaction(store),
                        "%r should not be joined to the transaction" % store)

    def test_wb_store_joins_transaction_on_register_event(self):
        """The Store joins the transaction when register-transaction
        is emitted.

        The Store tests check the various operations that trigger this
        event.
        """
        store = self.zstorm.get("name", "sqlite:")
        self.assertNotInTransaction(store)
        store._event.emit("register-transaction")
        self.assertInTransaction(store)

    def test_wb_store_joins_transaction_on_use_after_commit(self):
        store = self.zstorm.get("name", "sqlite:")
        store.execute("SELECT 1")
        transaction.commit()
        self.assertNotInTransaction(store)
        store.execute("SELECT 1")
        self.assertInTransaction(store)

    def test_wb_store_joins_transaction_on_use_after_abort(self):
        store = self.zstorm.get("name", "sqlite:")
        store.execute("SELECT 1")
        transaction.abort()
        self.assertNotInTransaction(store)
        store.execute("SELECT 1")
        self.assertInTransaction(store)

    def test_wb_store_joins_transaction_on_use_after_tpc_commit(self):
        """
        A store used after a two-phase commit re-joins the new transaction.
        """
        self.zstorm.set_default_uri("name", "sqlite:")
        self.zstorm.set_default_tpc("name", True)
        store = self.zstorm.get("name")
        store.begin = lambda xid: None
        store.prepare = lambda: None
        store.commit = lambda: None
        store.execute("SELECT 1")
        transaction.commit()
        self.assertNotInTransaction(store)
        store.execute("SELECT 1")
        self.assertInTransaction(store)

    def test_wb_store_joins_transaction_on_use_after_tpc_abort(self):
        """
        A store used after a rollback during a two-phase commit re-joins the
        new transaction.
        """
        self.zstorm.set_default_uri("name", "sqlite:")
        self.zstorm.set_default_tpc("name", True)
        store = self.zstorm.get("name")
        store.begin = lambda xid: None
        store.prepare = lambda: None
        store.rollback = lambda: None
        store.execute("SELECT 1")
        transaction.abort()
        self.assertNotInTransaction(store)
        store.execute("SELECT 1")
        self.assertInTransaction(store)

    def test_remove(self):
        removed_store = self.zstorm.get("name", "sqlite:")
        self.zstorm.remove(removed_store)
        for name, store in self.zstorm.iterstores():
            self.assertNotEqual(store, removed_store)
        self.assertRaises(ZStormError, self.zstorm.get, "name")

    def test_wb_removed_store_does_not_join_transaction(self):
        """If a store has been removed, it will not join the transaction."""
        store = self.zstorm.get("name", "sqlite:")
        self.zstorm.remove(store)
        store.execute("SELECT 1")
        self.assertNotInTransaction(store)

    def test_wb_removed_store_does_not_join_future_transactions(self):
        """If a store has been removed after joining a transaction, it
        will not join new transactions."""
        store = self.zstorm.get("name", "sqlite:")
        store.execute("SELECT 1")
        self.zstorm.remove(store)
        self.assertInTransaction(store)

        transaction.abort()
        store.execute("SELECT 1")
        self.assertNotInTransaction(store)

    def test_wb_cross_thread_store_does_not_join_transaction(self):
        """If a zstorm registered thread crosses over to another thread,
        it will not be usable."""
        store = self.zstorm.get("name", "sqlite:")

        failures = []
        def f():
            # We perform this twice to show that ZStormError is raised
            # consistently (i.e. not just the first time).
            for i in range(2):
                try:
                    store.execute("SELECT 1")
                except ZStormError:
                    failures.append("ZStormError raised")
                except Exception as exc:
                    failures.append("Expected ZStormError, got %r" % exc)
                else:
                    failures.append("Expected ZStormError, nothing raised")
                if self._isInTransaction(store):
                    failures.append("store was joined to transaction")
        thread = threading.Thread(target=f)
        thread.start()
        thread.join()
        self.assertEqual(failures, ["ZStormError raised"] * 2)

    def test_wb_reset(self):
        """_reset is used to reset the zstorm utility between zope test runs.
        """
        store = self.zstorm.get("name", "sqlite:")
        self.zstorm._reset()
        self.assertEqual(list(self.zstorm.iterstores()), [])

    def test_store_strong_reference(self):
        """
        The zstorm utility should be a strong reference to named stores so that
        it doesn't recreate stores uselessly.
        """
        store = self.zstorm.get("name", "sqlite:")
        store_ref = weakref.ref(store)
        transaction.abort()
        del store
        gc.collect()
        self.assertNotIdentical(store_ref(), None)
        store = self.zstorm.get("name")
        self.assertIdentical(store_ref(), store)


class ZStormUtilityTest(TestHelper):

    def is_supported(self):
        return has_transaction and has_zope_component

    def test_utility(self):
        provideUtility(ZStorm())
        self.assertTrue(isinstance(getUtility(IZStorm), ZStorm))




#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import os
import gc

from storm.database import create_database
from storm.properties import Enum, Int, List
from storm.info import get_obj_info
from storm.tests.store.base import StoreTest, EmptyResultSetTest, Foo
from storm.tests.helper import TestHelper


class Lst1:
    __storm_table__ = "lst1"
    id = Int(primary=True)
    ints = List(type=Int())

class LstEnum:
    __storm_table__ = "lst1"
    id = Int(primary=True)
    ints = List(type=Enum(map={"one": 1, "two": 2, "three": 3}))

class Lst2:
    __storm_table__ = "lst2"
    id = Int(primary=True)
    ints = List(type=List(type=Int()))

class FooWithSchema(Foo):
    __storm_table__ = "public.foo"


class PostgresStoreTest(TestHelper, StoreTest):

    def setUp(self):
        TestHelper.setUp(self)
        StoreTest.setUp(self)

    def tearDown(self):
        TestHelper.tearDown(self)
        StoreTest.tearDown(self)

    def is_supported(self):
        return bool(os.environ.get("STORM_POSTGRES_URI"))

    def create_database(self):
        self.database = create_database(os.environ["STORM_POSTGRES_URI"])

    def create_tables(self):
        connection = self.connection
        connection.execute("CREATE TABLE foo "
                           "(id SERIAL PRIMARY KEY,"
                           " title VARCHAR DEFAULT 'Default Title')")
        # Prevent dynamically created Foos from having conflicting ids.
        connection.execute("SELECT setval('foo_id_seq', 1000)")
        connection.execute("CREATE TABLE bar "
                           "(id SERIAL PRIMARY KEY,"
                           " foo_id INTEGER, title VARCHAR)")
        connection.execute("CREATE TABLE bin "
                           "(id SERIAL PRIMARY KEY, bin BYTEA, foo_id INTEGER)")
        connection.execute("CREATE TABLE link "
                           "(foo_id INTEGER, bar_id INTEGER,"
                           " PRIMARY KEY (foo_id, bar_id))")
        connection.execute("CREATE TABLE money "
                           "(id SERIAL PRIMARY KEY, value NUMERIC(6,4))")
        connection.execute("CREATE TABLE selfref "
                           "(id SERIAL PRIMARY KEY, title VARCHAR,"
                           " selfref_id INTEGER REFERENCES selfref(id))")
        connection.execute("CREATE TABLE lst1 "
                           "(id SERIAL PRIMARY KEY, ints INTEGER[])")
        connection.execute("CREATE TABLE lst2 "
                           "(id SERIAL PRIMARY KEY, ints INTEGER[][])")
        connection.execute("CREATE TABLE foovalue "
                           "(id SERIAL PRIMARY KEY, foo_id INTEGER,"
                           " value1 INTEGER, value2 INTEGER)")
        connection.execute("CREATE TABLE unique_id "
                           "(id UUID PRIMARY KEY)")
        connection.commit()

    def drop_tables(self):
        StoreTest.drop_tables(self)
        for table in ["lst1", "lst2"]:
            try:
                self.connection.execute("DROP TABLE %s" % table)
                self.connection.commit()
            except:
                self.connection.rollback()

    def test_list_variable(self):

        lst = Lst1()
        lst.id = 1
        lst.ints = [1,2,3,4]

        self.store.add(lst)

        result = self.store.execute("SELECT ints FROM lst1 WHERE id=1")
        self.assertEqual(result.get_one(), ([1,2,3,4],))

        del lst
        gc.collect()

        lst = self.store.find(Lst1, Lst1.ints == [1,2,3,4]).one()
        self.assertTrue(lst)

        lst.ints.append(5)

        result = self.store.execute("SELECT ints FROM lst1 WHERE id=1")
        self.assertEqual(result.get_one(), ([1,2,3,4,5],))

    def test_list_enum_variable(self):

        lst = LstEnum()
        lst.id = 1
        lst.ints = ["one", "two"]
        self.store.add(lst)

        result = self.store.execute("SELECT ints FROM lst1 WHERE id=1")
        self.assertEqual(result.get_one(), ([1,2],))

        del lst
        gc.collect()

        lst = self.store.find(LstEnum, LstEnum.ints == ["one", "two"]).one()
        self.assertTrue(lst)

        lst.ints.append("three")

        result = self.store.execute("SELECT ints FROM lst1 WHERE id=1")
        self.assertEqual(result.get_one(), ([1,2,3],))

    def test_list_variable_nested(self):

        lst = Lst2()
        lst.id = 1
        lst.ints = [[1, 2], [3, 4]]

        self.store.add(lst)

        result = self.store.execute("SELECT ints FROM lst2 WHERE id=1")
        self.assertEqual(result.get_one(), ([[1,2],[3,4]],))

        del lst
        gc.collect()

        lst = self.store.find(Lst2, Lst2.ints == [[1,2],[3,4]]).one()
        self.assertTrue(lst)

        lst.ints.append([5, 6])

        result = self.store.execute("SELECT ints FROM lst2 WHERE id=1")
        self.assertEqual(result.get_one(), ([[1,2],[3,4],[5,6]],))

    def test_add_find_with_schema(self):
        foo = FooWithSchema()
        foo.title = "Title"
        self.store.add(foo)
        self.store.flush()
        # We use find() here to actually exercise the backend code.
        # get() would just pick the object from the cache.
        self.assertEqual(self.store.find(FooWithSchema, id=foo.id).one(), foo)

    def test_wb_currval_based_identity(self):
        """
        Ensure that the currval()-based identity retrieval continues
        to work, even if we're currently running on a 8.2+ database.
        """
        self.database._version = 80109
        foo1 = self.store.add(Foo())
        self.store.flush()
        foo2 = self.store.add(Foo())
        self.store.flush()
        self.assertEqual(foo2.id-foo1.id, 1)

    def test_list_unnecessary_update(self):
        """
        Flushing an object with a list variable doesn't create an unnecessary
        UPDATE statement.
        """
        self.store.execute("INSERT INTO lst1 VALUES (1, '{}')", noresult=True)

        lst = self.store.find(Lst1, id=1).one()
        self.assertTrue(lst)
        self.store.invalidate()

        lst2 = self.store.find(Lst1, id=1).one()
        self.assertTrue(lst2)
        obj_info = get_obj_info(lst2)
        events = []
        obj_info.event.hook("changed", lambda *args: events.append(args))
        self.store.flush()
        self.assertEqual(events, [])


class PostgresEmptyResultSetTest(TestHelper, EmptyResultSetTest):

    def setUp(self):
        TestHelper.setUp(self)
        EmptyResultSetTest.setUp(self)

    def tearDown(self):
        TestHelper.tearDown(self)
        EmptyResultSetTest.tearDown(self)

    def is_supported(self):
        return bool(os.environ.get("STORM_POSTGRES_URI"))

    def create_database(self):
        self.database = create_database(os.environ["STORM_POSTGRES_URI"])

    def create_tables(self):
        self.connection.execute("CREATE TABLE foo "
                                "(id SERIAL PRIMARY KEY,"
                                " title VARCHAR DEFAULT 'Default Title')")
        self.connection.commit()



#
# Copyright (c) 2006, 2014 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Manage L{Schema}s across a set of L{Store} shards.

The L{Sharding} class can be used to perform schema operations (create,
upgrade, delete) against a set of L{Store}s. For example, let's say
we have two L{Schema}s and two L{Store}s we want to apply them to. We can
setup our L{Sharding} instance like this:

>>> schema1 = Schema(...)
>>> schema1 = Schema(...)
>>> store1 = Store(...)
>>> store2 = Store(...)
>>> sharding = Sharding()
>>> sharding.add(store1, schema1)
>>> sharding.add(store2, schema2)

And then perform schema maintenance operations across all shards:

>>> sharding.upgrade()

Patches will be applied "horizontally", meaning that the stores will always
be at the same patch level. See L{storm.schema.patch.PatchSet}.
"""

from storm.schema.schema import SchemaMissingError, UnappliedPatchesError


class PatchLevelMismatchError(Exception):
    """Raised when stores don't have all the same patch level."""


class Sharding:
    """Manage L{Shema}s over a collection of L{Store}s."""

    def __init__(self):
        self._stores = []  # Sequence of Stores with their Schemas

    def add(self, store, schema):
        """Add a new L{Store} shard.

        @param store: The L{Store} to add.
        @param schema: The L{Schema} the L{Store} is meant to have.
        """
        self._stores.append((store, schema))

    def create(self):
        """Create all schemas from scratch across all L{Store} shards."""
        for store, schema in self._stores:
            schema.create(store)

    def drop(self):
        """Drop all tables across all L{Store} shards."""
        for store, schema in self._stores:
            schema.drop(store)

    def delete(self):
        """Delete all table rows across all L{Store} shards."""
        for store, schema in self._stores:
            schema.delete(store)

    def upgrade(self):
        """Perform a schema upgrade.

        Pristine L{Store}s without any schema applied yet, will be initialized
        using L{Schema.create}.

        All other L{Store}s will be upgraded to the latest version of their
        L{Schema}s by applying all pending patches.

        The patching strategy is "horizontal", meaning that patch numbering
        must be the same across all L{Schema}s, and all L{Store}s must be at
        the same patch number. For example if the common patch number is N,
        the upgrade will start by applying patch number N + 1 to all
        non-pristine stores, following the order in which the stores were
        added to the L{Sharding}. Then the upgrade will apply all patches
        with number N + 2, etc.
        """
        stores_to_upgrade = []
        unapplied_versions = []
        for store, schema in self._stores:
            try:
                schema.check(store)
            except SchemaMissingError:
                schema.create(store)
            except UnappliedPatchesError as error:
                if not unapplied_versions:
                    unapplied_versions = error.unapplied_versions
                elif unapplied_versions != error.unapplied_versions:
                    raise PatchLevelMismatchError(
                        "Some stores have different patch levels")
                stores_to_upgrade.append((store, schema))

        for version in unapplied_versions:
            for store, schema in stores_to_upgrade:
                schema.advance(store, version)



#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import weakref

from storm.exceptions import (
    ClassInfoError, FeatureError, NoStoreError, WrongStoreError)
from storm.store import Store, get_where_for_args, LostObjectError
from storm.variables import LazyValue
from storm.expr import (
    Select, Column, Exists, ComparableExpr, SuffixExpr, LeftJoin, Not, SQLRaw,
    compare_columns, compile)
from storm.info import get_cls_info, get_obj_info


__all__ = ["Reference", "ReferenceSet", "Proxy"]


class LazyAttribute:
    """
    This descriptor will call the named attribute builder to
    initialize the given attribute on first access.  It avoids
    having a test at every single place where the attribute is
    touched when lazy initialization is wanted, and prevents
    paying the price of a normal property when classes are
    seldomly instantiated (the case of references).
    """

    def __init__(self, attr, attr_builder):
        self._attr = attr
        self._attr_builder = attr_builder

    def __get__(self, obj, cls=None):
        getattr(obj, self._attr_builder)()
        return getattr(obj, self._attr)


class PendingReferenceValue(LazyValue):
    """Lazy value to be used as a marker for unflushed foreign keys.

    When a reference is set to an object which is still unflushed,
    the foreign key in the local object remains set to this value
    until the object is flushed.
    """

PendingReferenceValue = PendingReferenceValue()


class Reference:
    """Descriptor for one-to-one relationships.

    This is typically used when the class that it is being defined on
    has a foreign key onto another table::

        class OtherGuy(object):
            ...
            id = Int()

        class MyGuy(object):
            ...
            other_guy_id = Int()
            other_guy = Reference(other_guy_id, OtherGuy.id)

    but can also be used for backwards references, where OtherGuy's
    table has a foreign key onto the class that you want this property
    on::

        class OtherGuy(object):
            ...
            my_guy_id = Int() # in the database, a foreign key to my_guy.id

        class MyGuy(object):
            ...
            id = Int()
            other_guy = Reference(id, OtherGuy.my_guy_id, on_remote=True)

    In both cases, C{MyGuy().other_guy} will resolve to the
    C{OtherGuy} instance which is linked to it. In the first case, it
    will be the C{OtherGuy} instance whose C{id} is equivalent to the
    C{MyGuy}'s C{other_guy_id}; in the second, it'll be the
    C{OtherGuy} instance whose C{my_guy_id} is equivalent to the
    C{MyGuy}'s C{id}.

    Assigning to the property, for example with C{MyGuy().other_guy =
    OtherGuy()}, will link the objects and update either
    C{MyGuy.other_guy_id} or C{OtherGuy.my_guy_id} accordingly.

    String references may be used in place of L{storm.expr.Column} objects
    throughout, and are resolved to columns using L{PropertyResolver}.
    """

    # Must initialize _relation later because we don't want to resolve
    # string references at definition time, since classes refered to might
    # not be available yet.  Notice that this attribute is "public" to the
    # Proxy class and the SQLObject wrapper.  It's still underlined because
    # it's *NOT* part of the public API of Storm (we'll modify it without
    # warnings!).
    _relation = LazyAttribute("_relation", "_build_relation")

    def __init__(self, local_key, remote_key, on_remote=False):
        """
        Create a Reference property.

        @param local_key: The sibling column which is the foreign key
            onto C{remote_key}. (unless C{on_remote} is passed; see
            below).
        @param remote_key: The column on the referred-to object which
            will have the same value as that for C{local_key} when
            resolved on an instance.
        @param on_remote: If specified, then the reference is
            backwards: It is the C{remote_key} which is a foreign key
            onto C{local_key}.
        """
        self._local_key = local_key
        self._remote_key = remote_key
        self._on_remote = on_remote
        self._cls = None

    def __get__(self, local, cls=None):
        if local is not None:
            # Don't use local here, as it might be security proxied.
            local = get_obj_info(local).get_obj()

        if self._cls is None:
            self._cls = _find_descriptor_class(cls or local.__class__, self)

        if local is None:
            return self

        remote = self._relation.get_remote(local)
        if remote is not None:
            return remote

        if self._relation.local_variables_are_none(local):
            return None

        store = Store.of(local)
        if store is None:
            return None

        if self._relation.remote_key_is_primary:
            remote = store.get(self._relation.remote_cls,
                               self._relation.get_local_variables(local))
        else:
            where = self._relation.get_where_for_remote(local)
            result = store.find(self._relation.remote_cls, where)
            remote = result.one()

        if remote is not None:
            self._relation.link(local, remote)

        return remote

    def __set__(self, local, remote):
        # Don't use local here, as it might be security proxied or something.
        local = get_obj_info(local).get_obj()

        if self._cls is None:
            self._cls = _find_descriptor_class(local.__class__, self)

        if remote is None:
            if self._on_remote:
                remote = self.__get__(local)
                if remote is None:
                    return
            else:
                remote = self._relation.get_remote(local)
            if remote is None:
                remote_info = None
            else:
                remote_info = get_obj_info(remote)
            self._relation.unlink(get_obj_info(local), remote_info, True)
        else:
            # Don't use remote here, as it might be
            # security proxied or something.
            try:
                remote = get_obj_info(remote).get_obj()
            except ClassInfoError:
                pass # It might fail when remote is a tuple or a raw value.
            self._relation.link(local, remote, True)

    def _build_relation(self):
        resolver = PropertyResolver(self, self._cls)
        self._local_key = resolver.resolve(self._local_key)
        self._remote_key = resolver.resolve(self._remote_key)
        self._relation = Relation(self._local_key, self._remote_key,
                                  False, self._on_remote)

    def __eq__(self, other):
        return self._relation.get_where_for_local(other)

    def __ne__(self, other):
        return Not(self == other)

    __hash__ = object.__hash__


class ReferenceSet:
    """Descriptor for many-to-one and many-to-many reference sets.

    This is typically used when another class has a foreign key onto the
    class being defined, either directly (the many-to-one case) or via an
    intermediate table (the many-to-many case).  For instance::

        class Person(Storm):
            ...
            id = Int(primary=True)
            email_addresses = ReferenceSet("id", "EmailAddress.owner_id")

        class EmailAddress(Storm):
            ...
            owner_id = Int(name="owner", allow_none=False)
            owner = Reference(owner_id, "Person.id")

        class TeamMembership(Storm):
            ...
            person_id = Int(name="person", allow_none=False)
            person = Reference(person_id, "Person.id")
            team_id = Int(name="team", allow_none=False)
            team = Reference(team_id, "Team.id")

        class Team(Storm):
            ...
            id = Int(primary=True)
            members = ReferenceSet(
                "id", "TeamMembership.team_id",
                "TeamMembership.person_id", "Person.id",
                order_by="Person.name")

    In this case, C{Person().email_addresses} resolves to a
    L{BoundReferenceSet} of all the email addresses linked to that person (a
    many-to-one relationship), while C{Team().members} resolves to a
    L{BoundIndirectReferenceSet} of all the members of that team (a
    many-to-many relationship).  These can be used in a somewhat similar way
    to L{ResultSet <storm.store.ResultSet>} objects.

    String references may be used in place of L{storm.expr.Column} objects
    throughout, and are resolved to columns using L{PropertyResolver}.
    """

    # Must initialize later because we don't want to resolve string
    # references at definition time, since classes refered to might
    # not be available yet.
    _relation1 = LazyAttribute("_relation1", "_build_relations")
    _relation2 = LazyAttribute("_relation2", "_build_relations")
    _order_by = LazyAttribute("_order_by", "_build_relations")

    def __init__(self, local_key1, remote_key1,
                 remote_key2=None, local_key2=None, order_by=None):
        """
        @param local_key1: The sibling column which has the same value as
            that for C{remote_key1} when resolved on an instance.
        @param remote_key1: The column on the referring object (in the case
            of a many-to-one relation) or on the intermediate table (in the
            case of a many-to-many relation) which is the foreign key onto
            C{local_key1}.
        @param remote_key2: In the case of a many-to-many relation, the
            column on the intermediate table which is the foreign key onto
            C{local_key2}.
        @param local_key2: In the case of a many-to-many relation, the
            column on the referred-to object which has the same value as
            C{remote_key2} when resolved on an instance.
        @param order_by: If not C{None}, order the resolved
            L{BoundReferenceSet} or L{BoundIndirectReferenceSet} by these
            columns, as in L{storm.store.ResultSet.order_by}.
        """
        self._local_key1 = local_key1
        self._remote_key1 = remote_key1
        self._remote_key2 = remote_key2
        self._local_key2 = local_key2
        self._default_order_by = order_by
        self._cls = None

    def __get__(self, local, cls=None):
        if local is not None:
            # Don't use local here, as it might be security proxied.
            local = get_obj_info(local).get_obj()

        if self._cls is None:
            self._cls = _find_descriptor_class(cls or local.__class__, self)

        if local is None:
            return self

        #store = Store.of(local)
        #if store is None:
        #    return None

        if self._relation2 is None:
            return BoundReferenceSet(self._relation1, local, self._order_by)
        else:
            return BoundIndirectReferenceSet(self._relation1,
                                             self._relation2, local,
                                             self._order_by)

    def __set__(self, local, value):
        raise FeatureError("Assigning to ReferenceSets not supported")

    def _build_relations(self):
        resolver = PropertyResolver(self, self._cls)

        if self._default_order_by is not None:
            self._order_by = resolver.resolve(self._default_order_by)
        else:
            self._order_by = None

        self._local_key1 = resolver.resolve(self._local_key1)
        self._remote_key1 = resolver.resolve(self._remote_key1)
        self._relation1 = Relation(self._local_key1, self._remote_key1,
                                   True, True)

        if self._local_key2 and self._remote_key2:
            self._local_key2 = resolver.resolve(self._local_key2)
            self._remote_key2 = resolver.resolve(self._remote_key2)
            self._relation2 = Relation(self._local_key2, self._remote_key2,
                                       True, True)
        else:
            self._relation2 = None


class BoundReferenceSetBase:

    def find(self, *args, **kwargs):
        store = Store.of(self._local)
        if store is None:
            raise NoStoreError("Can't perform operation without a store")
        where = self._get_where_clause()
        result = store.find(self._target_cls, where, *args, **kwargs)
        if self._order_by is not None:
            result.order_by(*self._order_by)
        return result

    def __iter__(self):
        return self.find().__iter__()

    def __getitem__(self, index):
        return self.find().__getitem__(index)

    def __contains__(self, item):
        return item in self.find()

    def is_empty(self):
        return self.find().is_empty()

    def first(self, *args, **kwargs):
        return self.find(*args, **kwargs).first()

    def last(self, *args, **kwargs):
        return self.find(*args, **kwargs).last()

    def any(self, *args, **kwargs):
        return self.find(*args, **kwargs).any()

    def one(self, *args, **kwargs):
        return self.find(*args, **kwargs).one()

    def values(self, *columns):
        return self.find().values(*columns)

    def order_by(self, *args):
        return self.find().order_by(*args)

    def count(self):
        return self.find().count()


class BoundReferenceSet(BoundReferenceSetBase):
    """An instance of a many-to-one relation."""

    def __init__(self, relation, local, order_by):
        self._relation = relation
        self._local = local
        self._target_cls = self._relation.remote_cls
        self._order_by = order_by

    def _get_where_clause(self):
        return self._relation.get_where_for_remote(self._local)

    def clear(self, *args, **kwargs):
        set_kwargs = {}
        for remote_column in self._relation.remote_key:
            set_kwargs[remote_column.name] = None
        store = Store.of(self._local)
        if store is None:
            raise NoStoreError("Can't perform operation without a store")
        where = self._relation.get_where_for_remote(self._local)
        store.find(self._target_cls, where, *args, **kwargs).set(**set_kwargs)

    def add(self, remote):
        self._relation.link(self._local, remote, True)

    def remove(self, remote):
        self._relation.unlink(get_obj_info(self._local),
                              get_obj_info(remote), True)


class BoundIndirectReferenceSet(BoundReferenceSetBase):
    """An instance of a many-to-many relation."""

    def __init__(self, relation1, relation2, local, order_by):
        self._relation1 = relation1
        self._relation2 = relation2
        self._local = local
        self._order_by = order_by

        self._target_cls = relation2.local_cls
        self._link_cls = relation1.remote_cls

    def _get_where_clause(self):
        return (self._relation1.get_where_for_remote(self._local) &
                self._relation2.get_where_for_join())

    def clear(self, *args, **kwargs):
        store = Store.of(self._local)
        if store is None:
            raise NoStoreError("Can't perform operation without a store")
        where = self._relation1.get_where_for_remote(self._local)
        if args or kwargs:
            filter = get_where_for_args(args, kwargs, self._target_cls)
            join = self._relation2.get_where_for_join()
            table = get_cls_info(self._target_cls).table
            where &= Exists(Select(SQLRaw("*"), join & filter, tables=table))
        store.find(self._link_cls, where).remove()

    def add(self, remote):
        link = self._link_cls()
        self._relation1.link(self._local, link, True)
        # Don't use remote here, as it might be security proxied or something.
        remote = get_obj_info(remote).get_obj()
        self._relation2.link(remote, link, True)

    def remove(self, remote):
        store = Store.of(self._local)
        if store is None:
            raise NoStoreError("Can't perform operation without a store")
        # Don't use remote here, as it might be security proxied or something.
        remote = get_obj_info(remote).get_obj()
        where = (self._relation1.get_where_for_remote(self._local) &
                 self._relation2.get_where_for_remote(remote))
        store.find(self._link_cls, where).remove()


class Proxy(ComparableExpr):
    """Proxy exposes a referred object's column as a local column.

    For example::

      class Foo(object):
          bar_id = Int()
          bar = Reference(bar_id, Bar.id)
          bar_title = Proxy(bar, Bar.title)

    For most uses, C{Foo.bar_title} should behave as if it were
    a native property of C{Foo}.
    """

    class RemoteProp:
        """
        This descriptor will resolve and set the _remote_prop attribute
        when it's first used. It avoids having a test at every single
        place where the attribute is touched.
        """
        def __get__(self, obj, cls=None):
            resolver = PropertyResolver(obj, obj._cls)
            obj._remote_prop = resolver.resolve_one(obj._unresolved_prop)
            return obj._remote_prop

    _remote_prop = RemoteProp()

    def __init__(self, reference, remote_prop):
        self._reference = reference
        self._unresolved_prop = remote_prop
        self._cls = None

    def __get__(self, obj, cls=None):
        if self._cls is None:
            self._cls = _find_descriptor_class(cls, self)
        if obj is None:
            return self
        # Have you counted how many descriptors we're dealing with here? ;-)
        return self._remote_prop.__get__(self._reference.__get__(obj))

    def __set__(self, obj, value):
        return self._remote_prop.__set__(self._reference.__get__(obj), value)

    @property
    def variable_factory(self):
        return self._remote_prop.variable_factory

@compile.when(Proxy)
def compile_proxy(compile, proxy, state):
    # Inject the join between the table of the class holding the proxy
    # and the table of the class which is the target of the reference.
    left_join = LeftJoin(proxy._reference._relation.local_cls,
                         proxy._remote_prop.table,
                         proxy._reference._relation.get_where_for_join())
    state.auto_tables.append(left_join)

    # And compile the remote property normally.
    return compile(proxy._remote_prop, state)


class Relation:

    def __init__(self, local_key, remote_key, many, on_remote):
        assert type(local_key) is tuple and type(remote_key) is tuple

        self.local_key = local_key
        self.remote_key = remote_key

        self.local_cls = getattr(self.local_key[0], "cls", None)
        self.remote_cls = self.remote_key[0].cls
        self.remote_key_is_primary = False

        primary_key = get_cls_info(self.remote_cls).primary_key
        if len(primary_key) == len(self.remote_key):
            for column1, column2 in zip(self.remote_key, primary_key):
                if column1.name != column2.name:
                    break
            else:
                self.remote_key_is_primary = True

        self.many = many
        self.on_remote = on_remote

        # XXX These should probably be weak dictionaries.
        self._local_columns = {}
        self._remote_columns = {}

        self._l_to_r = {}
        self._r_to_l = {}

    def get_remote(self, local):
        """Return the remote object for this relation, using the local cache.

        If the object in the cache is invalidated, we validate it again to
        check if it's still in the database.
        """
        local_info = get_obj_info(local)
        try:
            obj = local_info[self]["remote"]
        except KeyError:
            return None
        remote_info = get_obj_info(obj)
        if remote_info.get("invalidated"):
            try:
                Store.of(obj)._validate_alive(remote_info)
            except LostObjectError:
                return None
        return obj

    def get_where_for_remote(self, local):
        """Generate a column comparison expression for reference properties.

        The returned expression may be used to find objects of the I{remote}
        type referring to C{local}.
        """
        local_variables = self.get_local_variables(local)
        for variable in local_variables:
            if not variable.is_defined():
                Store.of(local).flush()
                break
        return compare_columns(self.remote_key, local_variables)

    def get_where_for_local(self, other):
        """Generate a column comparison expression for reference properties.

        The returned expression may be used to find objects of the I{local}
        type referring to C{other}.

        It handles the following cases::

            Class.reference == obj
            Class.reference == obj.id
            Class.reference == (obj.id1, obj.id2)

        Where the right-hand side is the C{other} object given.
        """
        try:
            obj_info = get_obj_info(other)
        except ClassInfoError:
            if type(other) is not tuple:
                remote_variables = (other,)
            else:
                remote_variables = other
        else:
            # Don't use other here, as it might be
            # security proxied or something.
            other = get_obj_info(other).get_obj()
            remote_variables = self.get_remote_variables(other)
        return compare_columns(self.local_key, remote_variables)

    def get_where_for_join(self):
        return compare_columns(self.local_key, self.remote_key)

    def get_local_variables(self, local):
        local_info = get_obj_info(local)
        return tuple(local_info.variables[column]
                     for column in self._get_local_columns(local.__class__))

    def local_variables_are_none(self, local):
        """Return true if all variables of the local key have None values."""
        local_info = get_obj_info(local)
        for column in self._get_local_columns(local.__class__):
            if local_info.variables[column].get() is not None:
                return False
        return True

    def get_remote_variables(self, remote):
        remote_info = get_obj_info(remote)
        return tuple(remote_info.variables[column]
                     for column in self._get_remote_columns(remote.__class__))

    def link(self, local, remote, setting=False):
        """Link objects to represent their relation.

        @param local: Object representing the I{local} side of the reference.

        @param remote: Object representing the I{remote} side of the reference,
            or the actual value to be set as the local key.

        @param setting: Pass true when the relationship is being newly created.
        """
        local_info = get_obj_info(local)

        try:
            remote_info = get_obj_info(remote)
        except ClassInfoError:
            # Must be a plain key. Just set it.
            # XXX I guess this is broken if self.on_remote is True.
            local_variables = self.get_local_variables(local)
            if type(remote) is not tuple:
                remote = (remote,)
            assert len(remote) == len(local_variables)
            for variable, value in zip(local_variables, remote):
                variable.set(value)
            return

        local_store = Store.of(local)
        remote_store = Store.of(remote)

        if setting:
            if local_store is None:
                if remote_store is None:
                    local_info.event.hook("added", self._add_all, local_info)
                    remote_info.event.hook("added", self._add_all, local_info)
                else:
                    remote_store.add(local)
                    local_store = remote_store
            elif remote_store is None:
                local_store.add(remote)
            elif local_store is not remote_store:
                raise WrongStoreError("%r and %r cannot be linked because they "
                                      "are in different stores." %
                                      (local, remote))

        # In cases below, we maintain a reference to the remote object
        # to make sure it won't get deallocated while the link is active.
        relation_data = local_info.get(self)
        if self.many:
            if relation_data is None:
                relation_data = local_info[self] = {"remote":
                                                    {remote_info: remote}}
            else:
                relation_data["remote"][remote_info] = remote
        else:
            if relation_data is None:
                relation_data = local_info[self] = {"remote": remote}
            else:
                old_remote = relation_data.get("remote")
                if old_remote is not None:
                    self.unlink(local_info, get_obj_info(old_remote))
                relation_data["remote"] = remote

        if setting:
            local_vars = local_info.variables
            remote_vars = remote_info.variables
            pairs = zip(self._get_local_columns(local.__class__),
                        self.remote_key)
            if self.on_remote:
                local_has_changed = False
                for local_column, remote_column in pairs:
                    local_var = local_vars[local_column]
                    if not local_var.is_defined():
                        remote_vars[remote_column].set(PendingReferenceValue)
                    else:
                        remote_vars[remote_column].set(local_var.get())
                    if local_var.has_changed():
                        local_has_changed = True

                if local_has_changed:
                    self._add_flush_order(local_info, remote_info)

                local_info.event.hook("changed", self._track_local_changes,
                                      remote_info)
                local_info.event.hook("flushed", self._break_on_local_flushed,
                                      remote_info)
                #local_info.event.hook("removed", self._break_on_local_removed,
                #                      remote_info)
                remote_info.event.hook("removed", self._break_on_remote_removed,
                                       weakref.ref(local_info))
            else:
                remote_has_changed = False
                for local_column, remote_column in pairs:
                    remote_var = remote_vars[remote_column]
                    if not remote_var.is_defined():
                        local_vars[local_column].set(PendingReferenceValue)
                    else:
                        local_vars[local_column].set(remote_var.get())
                    if remote_var.has_changed():
                        remote_has_changed = True

                if remote_has_changed:
                    self._add_flush_order(local_info, remote_info,
                                          remote_first=True)

                remote_info.event.hook("changed", self._track_remote_changes,
                                       local_info)
                remote_info.event.hook("flushed", self._break_on_remote_flushed,
                                       local_info)
                #local_info.event.hook("removed", self._break_on_remote_removed,
                #                      local_info)

                local_info.event.hook("changed", self._break_on_local_diverged,
                                      remote_info)
        else:
            local_info.event.hook("changed", self._break_on_local_diverged,
                                  remote_info)
            remote_info.event.hook("changed", self._break_on_remote_diverged,
                                   weakref.ref(local_info))
            if self.on_remote:
                remote_info.event.hook("removed", self._break_on_remote_removed,
                                       weakref.ref(local_info))

    def unlink(self, local_info, remote_info, setting=False):
        """Break the relation between the local and remote objects.

        @param setting: If true objects will be changed to persist breakage.
        """
        unhook = False
        relation_data = local_info.get(self)
        if relation_data is not None:
            if self.many:
                remote_infos = relation_data["remote"]
                if remote_info in remote_infos:
                    remote_infos.pop(remote_info, None)
                    unhook = True
            else:
                if relation_data.pop("remote", None) is not None:
                    unhook = True

        if unhook:
            local_store = Store.of(local_info)

            local_info.event.unhook("changed", self._track_local_changes,
                                    remote_info)
            local_info.event.unhook("changed", self._break_on_local_diverged,
                                    remote_info)
            local_info.event.unhook("flushed", self._break_on_local_flushed,
                                    remote_info)

            remote_info.event.unhook("changed", self._track_remote_changes,
                                     local_info)
            remote_info.event.unhook("changed", self._break_on_remote_diverged,
                                     weakref.ref(local_info))
            remote_info.event.unhook("flushed", self._break_on_remote_flushed,
                                     local_info)
            remote_info.event.unhook("removed", self._break_on_remote_removed,
                                     weakref.ref(local_info))

            if local_store is None:
                if not self.many or not remote_infos:
                    local_info.event.unhook("added", self._add_all, local_info)
                remote_info.event.unhook("added", self._add_all, local_info)
            else:
                flush_order = relation_data.get("flush_order")
                if flush_order is not None and remote_info in flush_order:
                    if self.on_remote:
                        local_store.remove_flush_order(local_info, remote_info)
                    else:
                        local_store.remove_flush_order(remote_info, local_info)
                    flush_order.remove(remote_info)

        if setting:
            if self.on_remote:
                remote_vars = remote_info.variables
                for remote_column in self.remote_key:
                    remote_vars[remote_column].set(None)
            else:
                local_vars = local_info.variables
                local_cols = self._get_local_columns(local_info.cls_info.cls)
                for local_column in local_cols:
                    local_vars[local_column].set(None)

    def _add_flush_order(self, local_info, remote_info, remote_first=False):
        """Tell the Store to flush objects in the specified order.

        We need to conditionally remove the flush order in unlink() only
        if we added it here.  Note that we can't just check if the Store
        has ordering on the (local, remote) pair, since it may have more
        than one request for ordering it, from different relations.

        @param local_info: The object info for the local object.
        @param remote_info: The object info for the remote object.
        @param remote_first: If True, remote_info will be flushed
                             before local_info.
        """
        local_store = Store.of(local_info)
        if local_store is not None:
            flush_order = local_info[self].setdefault("flush_order", set())
            if remote_info not in flush_order:
                flush_order.add(remote_info)
                if remote_first:
                    local_store.add_flush_order(remote_info, local_info)
                else:
                    local_store.add_flush_order(local_info, remote_info)

    def _track_local_changes(self, local_info, local_variable,
                             old_value, new_value, fromdb, remote_info):
        """Deliver changes in local to remote.

        This hook ensures that the remote object will keep track of
        changes done in the local object, either manually or at
        flushing time.
        """
        remote_column = self._get_remote_column(local_info.cls_info.cls,
                                                local_variable.column)
        if remote_column is not None:
            remote_info.variables[remote_column].set(new_value)
            self._add_flush_order(local_info, remote_info)

    def _track_remote_changes(self, remote_info, remote_variable,
                              old_value, new_value, fromdb, local_info):
        """Deliver changes in remote to local.

        This hook ensures that the local object will keep track of
        changes done in the remote object, either manually or at
        flushing time.
        """
        local_column = self._get_local_column(local_info.cls_info.cls,
                                              remote_variable.column)
        if local_column is not None:
            local_info.variables[local_column].set(new_value)
            self._add_flush_order(local_info, remote_info, remote_first=True)

    def _break_on_local_diverged(self, local_info, local_variable,
                                 old_value, new_value, fromdb, remote_info):
        """Break the remote/local relationship on diverging changes.

        This hook ensures that if the local object has an attribute
        changed by hand in a way that diverges from the remote object,
        it stops tracking changes.
        """
        remote_column = self._get_remote_column(local_info.cls_info.cls,
                                                local_variable.column)
        if remote_column is not None:
            variable = remote_info.variables[remote_column]
            if variable.get_lazy() is None and variable.get() != new_value:
                self.unlink(local_info, remote_info)

    def _break_on_remote_diverged(self, remote_info, remote_variable,
                                  old_value, new_value, fromdb, local_info_ref):
        """Break the remote/local relationship on diverging changes.

        This hook ensures that if the remote object has an attribute
        changed by hand in a way that diverges from the local object,
        the relationship is undone.
        """
        local_info = local_info_ref()
        if local_info is None:
            return
        local_column = self._get_local_column(local_info.cls_info.cls,
                                              remote_variable.column)
        if local_column is not None:
            local_value = local_info.variables[local_column].get()
            if local_value != new_value:
                self.unlink(local_info, remote_info)

    def _break_on_local_flushed(self, local_info, remote_info):
        """Break the remote/local relationship on flush."""
        self.unlink(local_info, remote_info)

    def _break_on_remote_flushed(self, remote_info, local_info):
        """Break the remote/local relationship on flush."""
        self.unlink(local_info, remote_info)

    def _break_on_remote_removed(self, remote_info, local_info_ref):
        """Break the remote relationship when the remote object is removed."""
        local_info = local_info_ref()
        if local_info is not None:
            self.unlink(local_info, remote_info)

    def _add_all(self, obj_info, local_info):
        store = Store.of(obj_info)
        store.add(local_info)
        local_info.event.unhook("added", self._add_all, local_info)

        def add(remote_info):
            remote_info.event.unhook("added", self._add_all, local_info)
            store.add(remote_info)
            self._add_flush_order(local_info, remote_info,
                                  remote_first=(not self.on_remote))

        if self.many:
            for remote_info in local_info[self]["remote"]:
                add(remote_info)
        else:
            add(get_obj_info(local_info[self]["remote"]))

    def _get_remote_columns(self, remote_cls):
        try:
            return self._remote_columns[remote_cls]
        except KeyError:
            columns = tuple(prop.__get__(None, remote_cls)
                            for prop in self.remote_key)
            self._remote_columns[remote_cls] = columns
            return columns

    def _get_local_columns(self, local_cls):
        try:
            return self._local_columns[local_cls]
        except KeyError:
            columns = tuple(prop.__get__(None, local_cls)
                            for prop in self.local_key)
            self._local_columns[local_cls] = columns
            return columns

    def _get_remote_column(self, local_cls, local_column):
        try:
            return self._l_to_r[local_cls].get(local_column)
        except KeyError:
            map = {}
            for local_prop, _remote_column in zip(self.local_key,
                                                  self.remote_key):
                map[local_prop.__get__(None, local_cls)] = _remote_column
            return self._l_to_r.setdefault(local_cls, map).get(local_column)

    def _get_local_column(self, local_cls, remote_column):
        try:
            return self._r_to_l[local_cls].get(remote_column)
        except KeyError:
            map = {}
            for local_prop, _remote_column in zip(self.local_key,
                                                   self.remote_key):
                map[_remote_column] = local_prop.__get__(None, local_cls)
            return self._r_to_l.setdefault(local_cls, map).get(remote_column)


class PropertyResolver:
    """Transform strings and pure properties (non-columns) into columns."""

    def __init__(self, reference, used_cls):
        self._reference = reference
        self._used_cls = used_cls

        self._registry = None
        self._namespace = None

    def resolve(self, properties):
        if not type(properties) is tuple:
            return (self.resolve_one(properties),)
        return tuple(self.resolve_one(property) for property in properties)

    def resolve_one(self, property):
        if type(property) is tuple:
            return self.resolve(property)
        elif isinstance(property, str):
            return self._resolve_string(property)
        elif isinstance(property, SuffixExpr):
            # XXX This covers cases like order_by=Desc("Bar.id"), see #620369.
            # Eventually we might want to add support for other types of
            # expressions
            property.expr = self.resolve(property.expr)
            return property
        elif not isinstance(property, Column):
            return _find_descriptor_obj(self._used_cls, property)
        return property

    def _resolve_string(self, property_path):
        if self._registry is None:
            try:
                self._registry = self._used_cls._storm_property_registry
            except AttributeError:
                raise RuntimeError("When using strings on references, "
                                   "classes involved must be subclasses "
                                   "of 'Storm'")
            cls = _find_descriptor_class(self._used_cls, self._reference)
            self._namespace = "%s.%s" % (cls.__module__, cls.__name__)

        return self._registry.get(property_path, self._namespace)


def _find_descriptor_class(used_cls, descr):
    for cls in used_cls.__mro__:
        for attr, _descr in cls.__dict__.items():
            if _descr is descr:
                return cls
    raise RuntimeError("Reference used in an unknown class")

def _find_descriptor_obj(used_cls, descr):
    for cls in used_cls.__mro__:
        for attr, _descr in cls.__dict__.items():
            if _descr is descr:
                return getattr(cls, attr)
    raise RuntimeError("Reference used in an unknown class")



from storm.tests import has_psycopg
from storm.tests.helper import TestHelper
from storm.tests.zope import has_transaction, has_zope_component
from storm.tests.twisted import has_twisted

if has_transaction and has_zope_component and has_twisted:
    import transaction

    from twisted.trial.unittest import TestCase
    from zope.component import getUtility

    from storm.zope.interfaces import IZStorm
    from storm.exceptions import IntegrityError, DisconnectionError

    from storm.twisted.transact import Transactor, transact
    from storm.twisted.testing import FakeThreadPool
else:
    # We can't use trial's TestCase as base
    TestCase = TestHelper
    TestHelper = object

if has_psycopg:
    from psycopg2.extensions import TransactionRollbackError


class TransactorTest(TestCase, TestHelper):

    def is_supported(self):
        return has_transaction and has_zope_component and has_twisted

    def setUp(self):
        TestCase.setUp(self)
        TestHelper.setUp(self)
        self.threadpool = FakeThreadPool()
        self.transaction = self.mocker.mock()
        self.transactor = Transactor(self.threadpool, self.transaction)
        self.function = self.mocker.mock()

    def test_run(self):
        """
        L{Transactor.run} executes a function in a thread, commits
        the transaction and returns a deferred firing the function result.
        """
        self.mocker.order()
        self.expect(self.function(1, arg=2)).result(3)
        self.expect(self.transaction.commit())
        self.mocker.replay()
        deferred = self.transactor.run(self.function, 1, arg=2)
        deferred.addCallback(self.assertEqual, 3)
        return deferred

    def test_run_with_function_failure(self):
        """
        If the given function raises an error, then L{Transactor.run}
        aborts the transaction and re-raises the same error.
        """
        self.mocker.order()
        self.expect(self.function()).throw(ZeroDivisionError())
        self.expect(self.transaction.abort())
        self.mocker.replay()
        deferred = self.transactor.run(self.function)
        self.assertFailure(deferred, ZeroDivisionError)
        return deferred

    def test_run_with_disconnection_error(self):
        """
        If the given function raises a L{DisconnectionError}, then a C{SELECT
        1} will be executed in each registered store such that C{psycopg}
        actually detects the disconnection.
        """
        self.transactor.retries = 0
        self.mocker.order()
        zstorm = self.mocker.mock()
        store1 = self.mocker.mock()
        store2 = self.mocker.mock()
        gu = self.mocker.replace(getUtility)
        self.expect(self.function()).throw(DisconnectionError())
        self.expect(gu(IZStorm)).result(zstorm)
        self.expect(zstorm.iterstores()).result(iter((("store1", store1),
                                                      ("store2", store2))))
        self.expect(store1.execute("SELECT 1"))
        self.expect(store2.execute("SELECT 1"))
        self.expect(self.transaction.abort())
        self.mocker.replay()
        deferred = self.transactor.run(self.function)
        self.assertFailure(deferred, DisconnectionError)
        return deferred

    def test_run_with_disconnection_error_in_execute_is_ignored(self):
        """
        If the given function raises a L{DisconnectionError}, then a C{SELECT
        1} will be executed in each registered store such that C{psycopg}
        actually detects the disconnection. If another L{DisconnectionError}
        happens during C{execute}, then it is ignored.
        """
        self.transactor.retries = 0
        zstorm = self.mocker.mock()
        store1 = self.mocker.mock()
        store2 = self.mocker.mock()
        gu = self.mocker.replace(getUtility)
        self.mocker.order()
        self.expect(self.function()).throw(DisconnectionError())
        self.expect(gu(IZStorm)).result(zstorm)
        self.expect(zstorm.iterstores()).result(iter((("store1", store1),
                                                      ("store2", store2))))
        self.expect(store1.execute("SELECT 1")).throw(DisconnectionError())
        self.expect(store2.execute("SELECT 1"))
        self.expect(self.transaction.abort())
        self.mocker.replay()
        deferred = self.transactor.run(self.function)
        self.assertFailure(deferred, DisconnectionError)
        return deferred

    def test_run_with_commit_failure(self):
        """
        If the given function succeeds but the transaction fails to commit,
        then L{Transactor.run} aborts the transaction and re-raises
        the commit exception.
        """
        self.mocker.order()
        self.expect(self.function())
        self.expect(self.transaction.commit()).throw(ZeroDivisionError())
        self.expect(self.transaction.abort())
        self.mocker.replay()
        deferred = self.transactor.run(self.function)
        self.assertFailure(deferred, ZeroDivisionError)
        return deferred

    def test_wb_default_transaction(self):
        """
        By default L{Transact} uses the global transaction manager.
        """
        transactor = Transactor(self.threadpool)
        self.assertIdentical(transaction, transactor._transaction)

    def test_decorate(self):
        """
        A L{transact} decorator can be used with methods of an object that
        contains a L{Transactor} instance as a C{transactor} instance variable,
        ensuring that the decorated function is called via L{Transactor.run}.
        """
        self.mocker.order()
        self.expect(self.transaction.commit())
        self.mocker.replay()

        @transact
        def function(self):
            """docstring"""
            return "result"

        # Function metadata is copied to the wrapper.
        self.assertEqual("docstring", function.__doc__)
        deferred = function(self)
        deferred.addCallback(self.assertEqual, "result")
        return deferred

    def test_run_with_integrity_error_retries(self):
        """
        If the given function raises a L{IntegrityError}, then the function
        will be retried another two times before letting the exception bubble
        up.
        """
        self.transactor.sleep = self.mocker.mock()
        self.transactor.uniform = self.mocker.mock()
        self.mocker.order()

        self.expect(self.function()).throw(IntegrityError())
        self.expect(self.transaction.abort())
        self.expect(self.transactor.uniform(1, 2 ** 1)).result(1)
        self.expect(self.transactor.sleep(1))

        self.expect(self.function()).throw(IntegrityError())
        self.expect(self.transaction.abort())
        self.expect(self.transactor.uniform(1, 2 ** 2)).result(2)
        self.expect(self.transactor.sleep(2))

        self.expect(self.function()).throw(IntegrityError())
        self.expect(self.transaction.abort())
        self.mocker.replay()

        deferred = self.transactor.run(self.function)
        self.assertFailure(deferred, IntegrityError)
        return deferred

    def test_run_with_transaction_rollback_error_retries(self):
        """
        If the given function raises a L{TransactionRollbackError}, then the
        function will be retried another two times before letting the exception
        bubble up.
        """
        if not has_psycopg:
            return

        self.transactor.sleep = self.mocker.mock()
        self.transactor.uniform = self.mocker.mock()
        self.mocker.order()

        self.expect(self.function()).throw(TransactionRollbackError())
        self.expect(self.transaction.abort())
        self.expect(self.transactor.uniform(1, 2 ** 1)).result(1)
        self.expect(self.transactor.sleep(1))

        self.expect(self.function()).throw(TransactionRollbackError())
        self.expect(self.transaction.abort())
        self.expect(self.transactor.uniform(1, 2 ** 2)).result(2)
        self.expect(self.transactor.sleep(2))

        self.expect(self.function()).throw(TransactionRollbackError())
        self.expect(self.transaction.abort())
        self.mocker.replay()

        deferred = self.transactor.run(self.function)
        self.assertFailure(deferred, TransactionRollbackError)
        return deferred

    def test_run_with_disconnection_error_retries(self):
        """
        If the given function raises a L{DisconnectionError}, then the
        function will be retried another two times before letting the exception
        bubble up.
        """
        zstorm = self.mocker.mock()
        gu = self.mocker.replace(getUtility)
        self.transactor.sleep = self.mocker.mock()
        self.transactor.uniform = self.mocker.mock()
        self.mocker.order()

        self.expect(self.function()).throw(DisconnectionError())
        self.expect(gu(IZStorm)).result(zstorm)
        self.expect(zstorm.iterstores()).result(iter(()))
        self.expect(self.transaction.abort())
        self.expect(self.transactor.uniform(1, 2 ** 1)).result(1)
        self.expect(self.transactor.sleep(1))

        self.expect(self.function()).throw(DisconnectionError())
        self.expect(gu(IZStorm)).result(zstorm)
        self.expect(zstorm.iterstores()).result(iter(()))
        self.expect(self.transaction.abort())
        self.expect(self.transactor.uniform(1, 2 ** 2)).result(2)
        self.expect(self.transactor.sleep(2))

        self.expect(self.function()).throw(DisconnectionError())
        self.expect(gu(IZStorm)).result(zstorm)
        self.expect(zstorm.iterstores()).result(iter(()))
        self.expect(self.transaction.abort())
        self.mocker.replay()

        deferred = self.transactor.run(self.function)
        self.assertFailure(deferred, DisconnectionError)
        return deferred

    def test_run_with_integrity_error_on_commit_retries(self):
        """
        If the given function raises a L{IntegrityError}, then the function
        will be retried another two times before letting the exception bubble
        up.
        """
        self.transactor.sleep = self.mocker.mock()
        self.transactor.uniform = self.mocker.mock()
        self.mocker.order()

        self.expect(self.function())
        self.expect(self.transaction.commit()).throw(IntegrityError())
        self.expect(self.transaction.abort())
        self.expect(self.transactor.uniform(1, 2 ** 1)).result(1)
        self.expect(self.transactor.sleep(1))

        self.expect(self.function())
        self.expect(self.transaction.commit()).throw(IntegrityError())
        self.expect(self.transaction.abort())
        self.expect(self.transactor.uniform(1, 2 ** 2)).result(2)
        self.expect(self.transactor.sleep(2))

        self.expect(self.function())
        self.expect(self.transaction.commit()).throw(IntegrityError())
        self.expect(self.transaction.abort())
        self.mocker.replay()

        deferred = self.transactor.run(self.function)
        self.assertFailure(deferred, IntegrityError)
        return deferred

    def test_run_with_on_retry_callback(self):
        """
        If a retry callback is passed with the C{on_retry} parameter, then
        it's invoked with the number of retries performed so far.
        """
        calls = []

        def on_retry(context):
            calls.append(context)

        self.transactor.on_retry = on_retry

        self.transactor.sleep = self.mocker.mock()
        self.transactor.uniform = self.mocker.mock()
        self.mocker.order()

        self.expect(self.function(1, a=2))
        error = IntegrityError()
        self.expect(self.transaction.commit()).throw(error)
        self.expect(self.transaction.abort())
        self.expect(self.transactor.uniform(1, 2 ** 1)).result(1)
        self.expect(self.transactor.sleep(1))

        self.expect(self.function(1, a=2))
        self.expect(self.transaction.commit())
        self.mocker.replay()

        deferred = self.transactor.run(self.function, 1, a=2)

        def check(_):
            [context] = calls

            self.assertEqual(self.function, context.function)
            self.assertEqual((1,), context.args)
            self.assertEqual({"a": 2}, context.kwargs)
            self.assertEqual(1, context.retry)
            self.assertEqual(1, context.retry)
            self.assertIs(error, context.error)

        return deferred.addCallback(check)



#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from storm.properties import PropertyPublisherMeta


__all__ = ["Storm"]


class Storm(metaclass=PropertyPublisherMeta):
    """An optional base class for objects stored in a Storm Store.

    It causes your subclasses to be associated with a Storm
    PropertyRegistry. It is necessary to use this if you want to
    specify References with strings.
    """




#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import os
import shutil

import transaction

from testresources import TestResourceManager
from zope.component import provideUtility, getUtility

from storm.schema.patch import UnknownPatchError
from storm.schema.sharding import Sharding
from storm.zope.zstorm import ZStorm, global_zstorm
from storm.zope.interfaces import IZStorm


class ZStormResourceManager(TestResourceManager):
    """Provide a L{ZStorm} resource to be used in test cases.

    The constructor is passed the details of the L{Store}s to be registered
    in the provided L{ZStore} resource. Then the C{make} and C{clean} methods
    make sure that such L{Store}s are properly setup and cleaned for each test.

    @param databases: A C{list} of C{dict}s holding the following keys:
        - 'name', the name of the store to be registered.
        - 'uri', the database URI to use to create the store.
        - 'schema', optionally, the L{Schema} for the tables in the store, if
          not given no schema will be applied.
        - 'schema-uri', optionally an alternate URI to use for applying the
          schema, if not given it defaults to 'uri'.

    @ivar force_delete: If C{True} for running L{Schema.delete} on a L{Store}
        even if no commit was performed by the test. Useful when running a test
        in a subprocess that might commit behind our back.
    @ivar use_global_zstorm: If C{True} then the C{global_zstorm} object from
        C{storm.zope.zstorm} will be used, instead of creating a new one. This
        is useful for code loading the zcml directives of C{storm.zope}.
    @ivar schema_stamp_dir: Optionally, a path to a directory that will be used
        to save timestamps of the schema's patch packages, so schema upgrades
        will be performed only when needed. This is just an optimisation to let
        the resource setup a bit faster.
    @ivar vertical_patching: If C{True}, patches will be applied "vertically",
        meaning that all patches for the first store will be applied, then
        all patches for the second store etc. Otherwise, if set to C{False}
        patches will be applied "horizontally" (see L{Sharding.upgrade}). The
        default is C{True} just because of backward-compatibility, but normally
        you should set it to C{False}.
    """
    force_delete = False
    use_global_zstorm = False
    schema_stamp_dir = None
    vertical_patching = True

    def __init__(self, databases):
        super().__init__()
        self._databases = databases
        self._zstorm = None
        self._schema_zstorm = None
        self._commits = {}
        self._schemas = {}
        self._sharding = []

    def make(self, dependencies):
        """Create a L{ZStorm} resource to be used by tests.

        @return: A L{ZStorm} object that will be shared among all tests using
            this resource manager.
        """
        if self._zstorm is None:

            if self.use_global_zstorm:
                self._zstorm = global_zstorm
            else:
                self._zstorm = ZStorm()
            self._schema_zstorm = ZStorm()

            databases = self._databases

            # Adapt the old databases format to the new one, for backward
            # compatibility. This should be eventually dropped.
            if isinstance(databases, dict):
                databases = [
                    {"name": name, "uri": uri, "schema": schema}
                    for name, (uri, schema) in databases.items()]

            # Provide the global IZStorm utility before applying patches, so
            # patch code can get the ztorm object if needed (e.g. looking up
            # other stores).
            provideUtility(self._zstorm)

            self._set_create_hook()

            enforce_schema = False
            for database in databases:
                name = database["name"]
                uri = database["uri"]
                schema = database.get("schema")
                schema_uri = database.get("schema-uri", uri)
                self._zstorm.set_default_uri(name, uri)
                if schema is not None:
                    # The configuration for this database does not include a
                    # schema definition, so we just setup the store (the user
                    # code should apply the schema elsewhere, if any)
                    self._schemas[name] = schema
                    self._schema_zstorm.set_default_uri(name, schema_uri)
                    schema.autocommit(False)
                    store = self._schema_zstorm.get(name)
                    if not self._sharding or self.vertical_patching:
                        self._sharding.append(Sharding())
                    sharding = self._sharding[-1]
                    sharding.add(store, schema)
                    if self._has_patch_package_changed(name, schema):
                        enforce_schema = True

            if enforce_schema:
                for sharding in self._sharding:
                    try:
                        sharding.upgrade()
                    except UnknownPatchError:
                        sharding.drop()
                        sharding.create()
                    except:
                        # An unknown error occured, let's drop all timestamps
                        # so subsequent runs won't assume that everything is
                        # fine
                        self._purge_schema_stamp_dir()
                        raise
                    else:
                        sharding.delete()

            # Commit all schema changes across all stores
            transaction.commit()

        elif getUtility(IZStorm) is not self._zstorm:
            # This probably means that the test code has overwritten our
            # utility, let's re-register it.
            provideUtility(self._zstorm)

        return self._zstorm

    def _set_create_hook(self):
        """
        Set a hook in ZStorm.create, so we can lazily set commit proxies.
        """
        self._zstorm.__real_create__ = self._zstorm.create

        def create_hook(name, uri=None):
            store = self._zstorm.__real_create__(name, uri=uri)
            if self._schemas.get(name) is not None:
                # Only set commit proxies for databases that have a schema
                # that we can use for cleanup
                self._set_commit_proxy(store)
            return store

        self._zstorm.create = create_hook

    def _set_commit_proxy(self, store):
        """Set a commit proxy to keep track of commits and clean up the tables.

        @param store: The L{Store} to set the commit proxy on. Any commit on
            this store will result in the associated tables to be cleaned upon
            tear down.
        """
        store.__real_commit__ = store.commit

        def commit_proxy():
            self._commits[store] = True
            store.__real_commit__()

        store.commit = commit_proxy

    def _has_patch_package_changed(self, name, schema):
        """Whether the schema for the given database is up-to-date.

        As an optimisation, if the C{schema_stamp_dir} attribute is set, then
        this method performs a fast check based on the patch directory
        timestamp rather than the database patch table, so connections and
        upgrade queries can be skipped if there's no need.

        @param name: The name of the database to check.
        @param schema: The schema to be ensured.

        @return: C{True} if the patch directory has changed and the schema
            needs to be updated, C{False} otherwise.
        """
        # If a schema stamp directory is set, then figure out whether there's
        # need to upgrade the schema by looking at timestamps.
        if self.schema_stamp_dir is not None:
            schema_mtime = self._get_schema_mtime(schema)
            schema_stamp_mtime = self._get_schema_stamp_mtime(name)

            # The modification time of the schema's patch directory matches our
            # timestamp, so the schema is already up-to-date
            if schema_mtime == schema_stamp_mtime:
                return False

            # Save the modification time of the schema's patch directory so in
            # subsequent runs we'll know if we're already up-to-date
            self._set_schema_stamp_mtime(name, schema_mtime)

        return True

    def _get_schema_mtime(self, schema):
        """
        Return the modification time of the C{schema}'s patch directory.
        """
        patch_directory = os.path.dirname(schema._patch_set._package.__file__)
        schema_stat = os.stat(patch_directory)
        return int(schema_stat.st_mtime)

    def _get_schema_stamp_mtime(self, name):
        """
        Return the modification time of schemas's patch directory, as saved
        in the stamp directory.
        """
        # Let's create the stamp directory if it doesn't exist
        if not os.path.exists(self.schema_stamp_dir):
            os.makedirs(self.schema_stamp_dir)

        schema_stamp_path = os.path.join(self.schema_stamp_dir, name)

        # Get the last schema modification time we ran the upgrade for, or -1
        # if this is our first run
        if os.path.exists(schema_stamp_path):
            with open(schema_stamp_path) as fd:
                schema_stamp_mtime = int(fd.read())
        else:
            schema_stamp_mtime = -1

        return schema_stamp_mtime

    def _set_schema_stamp_mtime(self, name, schema_mtime):
        """
        Save the schema's modification time in the stamp directory.
        """
        schema_stamp_path = os.path.join(self.schema_stamp_dir, name)
        with open(schema_stamp_path, "w") as fd:
            fd.write("%d" % schema_mtime)

    def _purge_schema_stamp_dir(self):
        """Remove the stamp directory."""
        if self.schema_stamp_dir and os.path.exists(self.schema_stamp_dir):
            shutil.rmtree(self.schema_stamp_dir)

    def clean(self, resource):
        """Clean up the stores after a test."""
        try:
            for name, store in self._zstorm.iterstores():
                # Ensure that the store is in a consistent state
                store.flush()
                # Clear the alive cache *before* abort is called,
                # to prevent a useless loop in Store.invalidate
                # over the alive objects
                store._alive.clear()
        finally:
            transaction.abort()

        # Clean up tables after each test if a commit was made
        needs_commit = False
        for name, store in self._zstorm.iterstores():
            if self.force_delete or store in self._commits:
                schema_store = self._schema_zstorm.get(name)
                schema = self._schemas[name]
                schema.delete(schema_store)
                needs_commit = True
        if needs_commit:
            transaction.commit()
        self._commits = {}



#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from io import StringIO
import logging
import shutil
import sys
import tempfile

from storm.tests import mocker


__all__ = ["TestHelper", "MakePath", "LogKeeper"]


class TestHelper(mocker.MockerTestCase):

    helpers = []

    def is_supported(self):
        return True

    def setUp(self):
        super().setUp()
        self._helper_instances = []
        for helper_factory in self.helpers:
            helper = helper_factory()
            helper.set_up(self)
            self._helper_instances.append(helper)

    def tearDown(self):
        for helper in reversed(self._helper_instances):
            helper.tear_down(self)
        super().tearDown()

    @property
    def _testMethod(self):
        return getattr(self, self._testMethodName)

    def run(self, result=None):
        # Skip if is_supported() does not return True.
        if not self.is_supported():
            if hasattr(result, "addSkip"):
                result.startTest(self)
                result.addSkip(self, "Test not supported")
            return
        super().run(result)

    def assertVariablesEqual(self, checked, expected):
        self.assertEqual(len(checked), len(expected))
        for check, expect in zip(checked, expected):
            self.assertEqual(check.__class__, expect.__class__)
            self.assertEqual(check.get(), expect.get())


class MakePath:

    def set_up(self, test_case):
        self.dirname = tempfile.mkdtemp()
        self.dirs = []
        self.counter = 0
        test_case.make_dir = self.make_dir
        test_case.make_path = self.make_path

    def tear_down(self, test_case):
        shutil.rmtree(self.dirname)
        [shutil.rmtree(dir) for dir in self.dirs]

    def make_dir(self):
        path = tempfile.mkdtemp()
        self.dirs.append(path)
        return path

    def make_path(self, content=None, path=None):
        if path is None:
            self.counter += 1
            path = "%s/%03d" % (self.dirname, self.counter)
        if content is not None:
            file = open(path, "w")
            try:
                file.write(content)
            finally:
                file.close()
        return path


class LogKeeper:
    """Record logging information.

    Puts a 'logfile' attribute on your test case, which is a StringIO
    containing all log output.
    """

    def set_up(self, test_case):
        logger = logging.getLogger()
        test_case.logfile = StringIO()
        handler = logging.StreamHandler(test_case.logfile)
        self.old_handlers = logger.handlers
        # Sanity check; this might not be 100% what we want
        if self.old_handlers:
            test_case.assertEqual(len(self.old_handlers), 1)
            test_case.assertEqual(self.old_handlers[0].stream, sys.stderr)
        logger.handlers = [handler]

    def tear_down(self, test_case):
        logging.getLogger().handlers = self.old_handlers



from twisted.python.failure import Failure
from twisted.internet.defer import execute

from storm.twisted.transact import Transactor


class FakeThreadPool:
    """
    A fake L{twisted.python.threadpool.ThreadPool}, running functions inside
    the main thread instead for easing tests.
    """

    def callInThreadWithCallback(self, onResult, func, *args, **kw):
        success = True
        try:
            result = func(*args, **kw)
        except:
            result = Failure()
            success = False

        onResult(success, result)


class FakeTransaction:

    def commit(self):
        pass

    def abort(self):
        pass


class FakeTransactor(Transactor):
    """
    A fake C{Transactor} wrapper that runs the given function in the main
    thread and performs basic checks on its return value.  If it has a
    C{__storm_table__} property a C{RuntimeError} is raised because Storm
    objects cannot be used outside the thread in which they were created.

    @seealso: L{Transactor}.
    """
    retries = 0
    on_retry = None

    sleep = lambda *args, **kwargs: None

    def __init__(self, transaction=None):
        if transaction is None:
            transaction = FakeTransaction()
        self._transaction = transaction

    def run(self, function, *args, **kwargs):
        deferred = execute(self._wrap, function, *args, **kwargs)
        return deferred.addCallback(self._check_result)

    def _check_result(self, result):
        if getattr(result, "__storm_table__", None) is not None:
            raise RuntimeError("Attempted to return a Storm object from a "
                               "transaction")
        return result



#!/usr/bin/env python3

import os
import re

from setuptools import setup, Extension, find_packages


if os.path.isfile("MANIFEST"):
    os.unlink("MANIFEST")


BUILD_CEXTENSIONS = True


with open("storm/__init__.py") as init_py:
    VERSION = re.search('version = "([^"]+)"', init_py.read()).group(1)


with open("README") as readme:
    long_description = readme.read()


tests_require = [
    "fixtures >= 1.3.0",
    "mysqlclient",
    "pgbouncer >= 0.0.7",
    "postgresfixture",
    "psycopg2 >= 2.3.0",
    "testresources >= 0.2.4",
    "testtools >= 0.9.8",
    "timeline >= 0.0.2",
    "transaction >= 1.0.0",
    "Twisted >= 10.0.0",
    "zope.component >= 3.8.0",
    "zope.configuration",
    "zope.interface >= 4.0.0",
    "zope.security >= 3.7.2",
    ]


setup(
    name="storm",
    version=VERSION,
    description=(
        "Storm is an object-relational mapper (ORM) for Python "
        "developed at Canonical."),
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author="Gustavo Niemeyer",
    author_email="gustavo@niemeyer.net",
    maintainer="Storm Developers",
    maintainer_email="storm@lists.canonical.com",
    license="LGPL",
    url="https://storm.canonical.com",
    download_url="https://launchpad.net/storm/+download",
    packages=find_packages(),
    package_data={"": ["*.zcml"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        ("License :: OSI Approved :: GNU Library or "
         "Lesser General Public License (LGPL)"),
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Database",
        "Topic :: Database :: Front-Ends",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    ext_modules=(BUILD_CEXTENSIONS and
                 [Extension("storm.cextensions", ["storm/cextensions.c"])]),
    # The following options are specific to setuptools but ignored (with a
    # warning) by distutils.
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.6",
    install_requires=["packaging >= 14.1"],
    test_suite="storm.tests.find_tests",
    tests_require=tests_require,
    extras_require={
        "doc": [
            "fixtures",  # so that storm.testing can be imported
            "sphinx",
            "sphinx-epytext",
            ],
        "test": tests_require,
        },
    )
